@@ -1,17 +1,18 @@
-/* Copyright (C) 2000-2003 MySQL AB
+/*
+   Copyright (c) 2000, 2010, Oracle and/or its affiliates. All rights reserved.
 
-  This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU General Public License as published by
-  the Free Software Foundation; version 2 of the License.
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; version 2 of the License.
 
-  This program is distributed in the hope that it will be useful,
-  but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-  GNU General Public License for more details.
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
 
-  You should have received a copy of the GNU General Public License
-  along with this program; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
 */
 
 /*
@@ -117,7 +118,8 @@ void ndb_serialize_cond(const Item *item, void *arg)
           if (item->type() == Item::FUNC_ITEM)
           {
             Item_func *func_item= (Item_func *) item;
-            if (func_item->functype() == Item_func::UNKNOWN_FUNC &&
+            if ((func_item->functype() == Item_func::UNKNOWN_FUNC ||
+                 func_item->functype() == Item_func::NEG_FUNC) &&
                 func_item->const_item())
             {
               // Skip any arguments since we will evaluate function instead
@@ -151,7 +153,7 @@ void ndb_serialize_cond(const Item *item, void *arg)
     }
     else
     {
-      Ndb_cond_stack *ndb_stack= context->stack_ptr;
+      Ndb_cond_stack *ndb_stack= context->cond_stack;
       Ndb_cond *prev_cond= context->cond_ptr;
       Ndb_cond *curr_cond= context->cond_ptr= new Ndb_cond();
       if (!ndb_stack->ndb_cond)
@@ -243,16 +245,19 @@ void ndb_serialize_cond(const Item *item, void *arg)
       {
         // End marker for condition group
         DBUG_PRINT("info", ("End of condition group"));
+        context->expect_no_length();
         curr_cond->ndb_item= new Ndb_item(NDB_END_COND);
       }
       else
       {
+        bool pop= TRUE;
+
         switch (item->type()) {
         case Item::FIELD_ITEM:
         {
           Item_field *field_item= (Item_field *) item;
           Field *field= field_item->field;
-          enum_field_types type= field->type();
+          enum_field_types type= field->real_type();
           /*
             Check that the field is part of the table of the handler
             instance and that we expect a field with of this result type.
@@ -263,17 +268,20 @@ void ndb_serialize_cond(const Item *item, void *arg)
             DBUG_PRINT("info", ("FIELD_ITEM"));
             DBUG_PRINT("info", ("table %s", tab->getName()));
             DBUG_PRINT("info", ("column %s", field->field_name));
-            DBUG_PRINT("info", ("type %d", field->type()));
+            DBUG_PRINT("info", ("column length %u", field->field_length));
+            DBUG_PRINT("info", ("type %d", field->real_type()));
             DBUG_PRINT("info", ("result type %d", field->result_type()));
-            
+
             // Check that we are expecting a field and with the correct
-            // result type
+            // result type and of length that can store the item value
             if (context->expecting(Item::FIELD_ITEM) &&
-                context->expecting_field_type(field->type()) &&
+                context->expecting_field_type(field->real_type()) &&
+                context->expecting_max_length(field->field_length) &&
                 (context->expecting_field_result(field->result_type()) ||
                  // Date and year can be written as string or int
                  ((type == MYSQL_TYPE_TIME ||
                    type == MYSQL_TYPE_DATE || 
+                   type == MYSQL_TYPE_NEWDATE || 
                    type == MYSQL_TYPE_YEAR ||
                    type == MYSQL_TYPE_DATETIME)
                   ? (context->expecting_field_result(STRING_RESULT) ||
@@ -297,6 +305,7 @@ void ndb_serialize_cond(const Item *item, void *arg)
                 // We have not seen second argument yet
                 if (type == MYSQL_TYPE_TIME ||
                     type == MYSQL_TYPE_DATE || 
+                    type == MYSQL_TYPE_NEWDATE || 
                     type == MYSQL_TYPE_YEAR ||
                     type == MYSQL_TYPE_DATETIME)
                 {
@@ -310,6 +319,7 @@ void ndb_serialize_cond(const Item *item, void *arg)
                     context->expect_only(Item::STRING_ITEM);
                     context->expect(Item::VARBIN_ITEM);
                     context->expect_collation(field_item->collation.collation);
+                    context->expect_max_length(field->field_length);
                     break;
                   case REAL_RESULT:
                     context->expect_only(Item::REAL_ITEM);
@@ -339,6 +349,7 @@ void ndb_serialize_cond(const Item *item, void *arg)
                     !context->expecting_collation(item->collation.collation)
                     && type != MYSQL_TYPE_TIME
                     && type != MYSQL_TYPE_DATE
+                    && type != MYSQL_TYPE_NEWDATE
                     && type != MYSQL_TYPE_YEAR
                     && type != MYSQL_TYPE_DATETIME)
                 {
@@ -369,8 +380,9 @@ void ndb_serialize_cond(const Item *item, void *arg)
         {
           Item_func *func_item= (Item_func *) item;
           // Check that we expect a function or functional expression here
-          if (context->expecting(Item::FUNC_ITEM) || 
-              func_item->functype() == Item_func::UNKNOWN_FUNC)
+          if (context->expecting(Item::FUNC_ITEM) ||
+              func_item->functype() == Item_func::UNKNOWN_FUNC ||
+              func_item->functype() == Item_func::NEG_FUNC)
             context->expect_nothing();
           else
           {
@@ -378,7 +390,7 @@ void ndb_serialize_cond(const Item *item, void *arg)
             context->supported= FALSE;
             break;
           }
-          
+          context->expect_no_length();
           switch (func_item->functype()) {
           case Item_func::EQ_FUNC:
           {
@@ -484,16 +496,27 @@ void ndb_serialize_cond(const Item *item, void *arg)
           }
           case Item_func::LIKE_FUNC:
           {
+            Ndb_expect_stack* expect_next= new Ndb_expect_stack();
             DBUG_PRINT("info", ("LIKE_FUNC"));      
             curr_cond->ndb_item= new Ndb_item(func_item->functype(),
                                               func_item);      
-            context->expect(Item::STRING_ITEM);
+
+            /*
+              Ndb currently only supports pushing
+              <field> LIKE <string> | <func>
+              we thus push <string> | <func>
+              on the expect stack to catch that we
+              don't support <string> LIKE <field>.
+             */
             context->expect(Item::FIELD_ITEM);
             context->expect_only_field_type(MYSQL_TYPE_STRING);
             context->expect_field_type(MYSQL_TYPE_VAR_STRING);
             context->expect_field_type(MYSQL_TYPE_VARCHAR);
             context->expect_field_result(STRING_RESULT);
-            context->expect(Item::FUNC_ITEM);
+            expect_next->expect(Item::STRING_ITEM);
+            expect_next->expect(Item::FUNC_ITEM);
+            context->expect_stack.push(expect_next);
+            pop= FALSE;
             break;
           }
           case Item_func::ISNULL_FUNC:
@@ -584,11 +607,30 @@ void ndb_serialize_cond(const Item *item, void *arg)
             context->expect(Item::FUNC_ITEM);
             break;
           }
+          case Item_func::NEG_FUNC:
           case Item_func::UNKNOWN_FUNC:
           {
-            DBUG_PRINT("info", ("UNKNOWN_FUNC %s", 
+            /*
+              Constant expressions of the type
+              -17, 1+2, concat(0xBB, '%') will
+              be evaluated before pushed.
+             */
+            DBUG_PRINT("info", ("Function %s", 
                                 func_item->const_item()?"const":""));  
             DBUG_PRINT("info", ("result type %d", func_item->result_type()));
+            /*
+              Check if we are rewriting queries of the type
+              <const> BETWEEN|IN <func> ...
+              as this is currently not supported.
+             */
+            if (context->rewrite_stack &&
+                context->rewrite_stack->left_hand_item &&
+                context->rewrite_stack->left_hand_item->type()
+                != Item::FIELD_ITEM)
+            {
+              DBUG_PRINT("info", ("Function during rewrite not supported"));
+              context->supported= FALSE;
+            }
             if (func_item->const_item())
             {
               switch (func_item->result_type()) {
@@ -710,7 +752,9 @@ void ndb_serialize_cond(const Item *item, void *arg)
         }
         case Item::STRING_ITEM:
           DBUG_PRINT("info", ("STRING_ITEM")); 
-          if (context->expecting(Item::STRING_ITEM)) 
+          // Check that we do support pushing the item value length
+          if (context->expecting(Item::STRING_ITEM) &&
+              context->expecting_length(item->max_length)) 
           {
 #ifndef DBUG_OFF
             char buff[256];
@@ -729,12 +773,14 @@ void ndb_serialize_cond(const Item *item, void *arg)
               context->expect_only(Item::FIELD_ITEM);
               context->expect_only_field_result(STRING_RESULT);
               context->expect_collation(item->collation.collation);
+              context->expect_length(item->max_length);
             }
             else 
             {
               // Expect another logical expression
               context->expect_only(Item::FUNC_ITEM);
               context->expect(Item::COND_ITEM);
+              context->expect_no_length();
               // Check that we are comparing with a field with same collation
               if (!context->expecting_collation(item->collation.collation))
               {
@@ -883,6 +929,8 @@ void ndb_serialize_cond(const Item *item, void *arg)
           context->supported= FALSE;
         }
         }
+        if (pop)
+          context->expect_stack.pop();
       }
       if (context->supported && context->rewrite_stack)
       {
@@ -896,6 +944,7 @@ void ndb_serialize_cond(const Item *item, void *arg)
           curr_cond= context->cond_ptr= new Ndb_cond();
           curr_cond->prev= prev_cond;
           prev_cond->next= curr_cond;
+          context->expect_no_length();
           curr_cond->ndb_item= new Ndb_item(NDB_END_COND);
           // Pop rewrite stack
           context->rewrite_stack=  rewrite_context->next;
@@ -1332,29 +1381,34 @@ ha_ndbcluster_cond::build_scan_filter(Ndb_cond * &cond, NdbScanFilter *filter)
 }
 
 int
-ha_ndbcluster_cond::generate_scan_filter(NdbScanOperation *op)
+ha_ndbcluster_cond::generate_scan_filter(NdbInterpretedCode* code,
+                                         NdbScanOperation::ScanOptions* options)
 {
   DBUG_ENTER("generate_scan_filter");
 
   if (m_cond_stack)
   {
-    NdbScanFilter filter(op, false); // don't abort on too large
+    NdbScanFilter filter(code);
     
-    int ret=generate_scan_filter_from_cond(filter);
+    int ret= generate_scan_filter_from_cond(filter);
     if (ret != 0)
     {
-      const NdbError& err=filter.getNdbError();
+      const NdbError& err= filter.getNdbError();
       if (err.code == NdbScanFilter::FilterTooLarge)
       {
         // err.message has static storage
         DBUG_PRINT("info", ("%s", err.message));
         push_warning(current_thd, MYSQL_ERROR::WARN_LEVEL_WARN,
                      err.code, err.message);
-        ret=0;
       }
+      else
+        DBUG_RETURN(ret);
+    }
+    else
+    {
+      options->interpretedCode= code;
+      options->optionsPresent|= NdbScanOperation::ScanOptions::SO_INTERPRETED;
     }
-    if (ret != 0)
-      DBUG_RETURN(ret);
   }
   else
   {  
@@ -1397,43 +1451,162 @@ ha_ndbcluster_cond::generate_scan_filter_from_cond(NdbScanFilter& filter)
 }
 
 
-int ha_ndbcluster_cond::generate_scan_filter_from_key(NdbScanOperation *op,
+/*
+  Optimizer sometimes does hash index lookup of a key where some
+  key parts are null.  The set of cases where this happens makes
+  no sense but cannot be ignored since optimizer may expect the result
+  to be filtered accordingly.  The scan is actually on the table and
+  the index bounds are pushed down.
+*/
+int ha_ndbcluster_cond::generate_scan_filter_from_key(NdbInterpretedCode* code,
+                                                      NdbScanOperation::ScanOptions* options,
                                                       const KEY* key_info, 
-                                                      const uchar *key, 
-                                                      uint key_len,
+                                                      const key_range *start_key,
+                                                      const key_range *end_key,
                                                       uchar *buf)
 {
-  KEY_PART_INFO* key_part= key_info->key_part;
-  KEY_PART_INFO* end= key_part+key_info->key_parts;
-  NdbScanFilter filter(op, true); // abort on too large
-  int res;
   DBUG_ENTER("generate_scan_filter_from_key");
 
-  filter.begin(NdbScanFilter::AND);
-  for (; key_part != end; key_part++) 
+#ifndef DBUG_OFF
   {
-    Field* field= key_part->field;
-    uint32 pack_len= field->pack_length();
-    const uchar* ptr= key;
-    DBUG_PRINT("info", ("Filtering value for %s", field->field_name));
-    DBUG_DUMP("key", ptr, pack_len);
-    if (key_part->null_bit)
+    DBUG_PRINT("info", ("key parts:%u length:%u",
+                        key_info->key_parts, key_info->key_length));
+    const key_range* keylist[2]={ start_key, end_key };
+    for (uint j=0; j <= 1; j++)
     {
-      DBUG_PRINT("info", ("Generating ISNULL filter"));
-      if (filter.isnull(key_part->fieldnr-1) == -1)
-	DBUG_RETURN(1);
+      char buf[8192];
+      const key_range* key=keylist[j];
+      if (key == 0)
+      {
+        sprintf(buf, "key range %u: none", j);
+      }
+      else
+      {
+        sprintf(buf, "key range %u: flag:%u part", j, key->flag);
+        const KEY_PART_INFO* key_part=key_info->key_part;
+        const uchar* ptr=key->key;
+        for (uint i=0; i < key_info->key_parts; i++)
+        {
+          sprintf(buf+strlen(buf), " %u:", i);
+          for (uint k=0; k < key_part->store_length; k++)
+          {
+            sprintf(buf+strlen(buf), " %02x", ptr[k]);
+          }
+          ptr+=key_part->store_length;
+          if (ptr - key->key >= (ptrdiff_t)key->length)
+          {
+            /*
+              key_range has no count of parts so must test byte length.
+              But this is not the place for following assert.
+            */
+            // DBUG_ASSERT(ptr - key->key == key->length);
+            break;
+          }
+          key_part++;
+        }
+      }
+      DBUG_PRINT("info", ("%s", buf));
     }
-    else
+  }
+#endif
+
+  NdbScanFilter filter(code);
+  int res;
+  filter.begin(NdbScanFilter::AND);
+  do
+  {
+    /*
+      Case "x is not null".
+      Seen with index(x) where it becomes range "null < x".
+      Not seen with index(x,y) for any combination of bounds
+      which include "is not null".
+    */
+    if (start_key != 0 &&
+        start_key->flag == HA_READ_AFTER_KEY &&
+        end_key == 0 &&
+        key_info->key_parts == 1)
     {
-      DBUG_PRINT("info", ("Generating EQ filter"));
-      if (filter.cmp(NdbScanFilter::COND_EQ, 
-		     key_part->fieldnr-1,
-		     ptr,
-		     pack_len) == -1)
-	DBUG_RETURN(1);
+      const KEY_PART_INFO* key_part=key_info->key_part;
+      if (key_part->null_bit != 0) // nullable (must be)
+      {
+        const uchar* ptr= start_key->key;
+        if (ptr[0] != 0) // null (in "null < x")
+        {
+          DBUG_PRINT("info", ("Generating ISNOTNULL filter for nullable %s",
+                              key_part->field->field_name));
+          if (filter.isnotnull(key_part->fieldnr-1) == -1)
+            DBUG_RETURN(1);
+          break;
+        }
+      }
     }
-    key += key_part->store_length;
-  }      
+
+    /*
+      Case "x is null" in an EQ range.
+      Seen with index(x) for "x is null".
+      Seen with index(x,y) for "x is null and y = 1".
+      Not seen with index(x,y) for "x is null and y is null".
+      Seen only when all key parts are present (but there is
+      no reason to limit the code to this case).
+    */
+    if (start_key != 0 &&
+        start_key->flag == HA_READ_KEY_EXACT &&
+        end_key != 0 &&
+        end_key->flag == HA_READ_AFTER_KEY &&
+        start_key->length == end_key->length &&
+        memcmp(start_key->key, end_key->key, start_key->length) == 0)
+    {
+      const KEY_PART_INFO* key_part=key_info->key_part;
+      const uchar* ptr=start_key->key;
+      for (uint i=0; i < key_info->key_parts; i++)
+      {
+        const Field* field=key_part->field;
+        if (key_part->null_bit) // nullable
+        {
+          if (ptr[0] != 0) // null
+          {
+            DBUG_PRINT("info", ("Generating ISNULL filter for nullable %s",
+                                field->field_name));
+            if (filter.isnull(key_part->fieldnr-1) == -1)
+              DBUG_RETURN(1);
+          }
+          else
+          {
+            DBUG_PRINT("info", ("Generating EQ filter for nullable %s",
+                                field->field_name));
+            if (filter.cmp(NdbScanFilter::COND_EQ, 
+                           key_part->fieldnr-1,
+                           ptr + 1, // skip null-indicator byte
+                           field->pack_length()) == -1)
+              DBUG_RETURN(1);
+          }
+        }
+        else
+        {
+          DBUG_PRINT("info", ("Generating EQ filter for non-nullable %s",
+                              field->field_name));
+          if (filter.cmp(NdbScanFilter::COND_EQ, 
+                         key_part->fieldnr-1,
+                         ptr,
+                         field->pack_length()) == -1)
+            DBUG_RETURN(1);
+        }
+        ptr+=key_part->store_length;
+        if (ptr - start_key->key >= (ptrdiff_t)start_key->length)
+        {
+          break;
+        }
+        key_part++;
+      }
+      break;
+    }
+
+    DBUG_PRINT("info", ("Unknown hash index scan"));
+    // enable to catch new cases when optimizer changes
+    // DBUG_ASSERT(false);
+  }
+  while (0);
+
   // Add any pushed condition
   if (m_cond_stack &&
       (res= generate_scan_filter_from_cond(filter)))
@@ -1442,6 +1615,9 @@ int ha_ndbcluster_cond::generate_scan_filter_from_key(NdbScanOperation *op,
   if (filter.end() == -1)
     DBUG_RETURN(1);
 
+  options->interpretedCode= code;
+  options->optionsPresent|= NdbScanOperation::ScanOptions::SO_INTERPRETED;
+
   DBUG_RETURN(0);
 }
 
