@@ -1,4 +1,4 @@
-/* Copyright (C) 2000-2006 MySQL AB
+/* Copyright (c) 2000, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +11,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 
 /*
@@ -116,7 +117,7 @@ static handler *myisammrg_create_handler(handlerton *hton,
 */
 
 ha_myisammrg::ha_myisammrg(handlerton *hton, TABLE_SHARE *table_arg)
-  :handler(hton, table_arg), file(0)
+  :handler(hton, table_arg), file(0), is_cloned(0)
 {}
 
 
@@ -137,14 +138,15 @@ extern int table2myisam(TABLE *table_arg, MI_KEYDEF **keydef_out,
 extern int check_definition(MI_KEYDEF *t1_keyinfo, MI_COLUMNDEF *t1_recinfo,
                             uint t1_keys, uint t1_recs,
                             MI_KEYDEF *t2_keyinfo, MI_COLUMNDEF *t2_recinfo,
-                            uint t2_keys, uint t2_recs, bool strict);
+                            uint t2_keys, uint t2_recs, bool strict,
+                            TABLE *table_arg);
 static void split_file_name(const char *file_name,
 			    LEX_STRING *db, LEX_STRING *name);
 
 
 extern "C" void myrg_print_wrong_table(const char *table_name)
 {
-  LEX_STRING db, name;
+  LEX_STRING db= {NULL, 0}, name;
   char buf[FN_REFLEN];
   split_file_name(table_name, &db, &name);
   memcpy(buf, db.str, db.length);
@@ -213,36 +215,14 @@ const char *ha_myisammrg::index_type(uint key_number)
 static int myisammrg_parent_open_callback(void *callback_param,
                                           const char *filename)
 {
-  ha_myisammrg  *ha_myrg;
-  TABLE         *parent;
+  ha_myisammrg  *ha_myrg= (ha_myisammrg*) callback_param;
+  TABLE         *parent= ha_myrg->table_ptr();
   TABLE_LIST    *child_l;
-  const char    *db;
-  const char    *table_name;
-  uint          dirlen;
+  size_t        dirlen;
   char          dir_path[FN_REFLEN];
+  char          name_buf[NAME_LEN];
   DBUG_ENTER("myisammrg_parent_open_callback");
 
-  /* Extract child table name and database name from filename. */
-  dirlen= dirname_length(filename);
-  if (dirlen >= FN_REFLEN)
-  {
-    /* purecov: begin inspected */
-    DBUG_PRINT("error", ("name too long: '%.64s'", filename));
-    my_errno= ENAMETOOLONG;
-    DBUG_RETURN(1);
-    /* purecov: end */
-  }
-  table_name= filename + dirlen;
-  dirlen--; /* Strip off trailing '/'. */
-  memcpy(dir_path, filename, dirlen);
-  dir_path[dirlen]= '\0';
-  db= base_name(dir_path);
-  dirlen-= db - dir_path; /* This is now the length of 'db'. */
-  DBUG_PRINT("myrg", ("open: '%s'.'%s'", db, table_name));
-
-  ha_myrg= (ha_myisammrg*) callback_param;
-  parent= ha_myrg->table_ptr();
-
   /* Get a TABLE_LIST object. */
   if (!(child_l= (TABLE_LIST*) alloc_root(&parent->mem_root,
                                           sizeof(TABLE_LIST))))
@@ -254,13 +234,70 @@ static int myisammrg_parent_open_callback(void *callback_param,
   }
   bzero((char*) child_l, sizeof(TABLE_LIST));
 
-  /* Set database (schema) name. */
-  child_l->db_length= dirlen;
-  child_l->db= strmake_root(&parent->mem_root, db, dirlen);
-  /* Set table name. */
-  child_l->table_name_length= strlen(table_name);
-  child_l->table_name= strmake_root(&parent->mem_root, table_name,
-                                    child_l->table_name_length);
+  /*
+    Depending on MySQL version, filename may be encoded by table name to
+    file name encoding or not. Always encoded if parent table is created
+    by 5.1.46+. Encoded if parent is created by 5.1.6+ and child table is
+    in different database.
+  */
+  if (!has_path(filename))
+  {
+    /* Child is in the same database as parent. */
+    child_l->db_length= parent->s->db.length;
+    child_l->db= strmake_root(&parent->mem_root, parent->s->db.str,
+                              child_l->db_length);
+    /* Child table name is encoded in parent dot-MRG starting with 5.1.46. */
+    if (parent->s->mysql_version >= 50146)
+    {
+      child_l->table_name_length= filename_to_tablename(filename, name_buf,
+                                                        sizeof(name_buf));
+      child_l->table_name= strmake_root(&parent->mem_root, name_buf,
+                                        child_l->table_name_length);
+    }
+    else
+    {
+      child_l->table_name_length= strlen(filename);
+      child_l->table_name= strmake_root(&parent->mem_root, filename,
+                                        child_l->table_name_length);
+    }
+  }
+  else
+  {
+    DBUG_ASSERT(strlen(filename) < sizeof(dir_path));
+    fn_format(dir_path, filename, "", "", 0);
+    /* Extract child table name and database name from filename. */
+    dirlen= dirname_length(dir_path);
+    /* Child db/table name is encoded in parent dot-MRG starting with 5.1.6. */
+    if (parent->s->mysql_version >= 50106)
+    {
+      child_l->table_name_length= filename_to_tablename(dir_path + dirlen,
+                                                        name_buf,
+                                                        sizeof(name_buf));
+      child_l->table_name= strmake_root(&parent->mem_root, name_buf,
+                                        child_l->table_name_length);
+      dir_path[dirlen - 1]= 0;
+      dirlen= dirname_length(dir_path);
+      child_l->db_length= filename_to_tablename(dir_path + dirlen, name_buf,
+                                                sizeof(name_buf));
+      child_l->db= strmake_root(&parent->mem_root, name_buf, child_l->db_length);
+    }
+    else
+    {
+      child_l->table_name_length= strlen(dir_path + dirlen);
+      child_l->table_name= strmake_root(&parent->mem_root, dir_path + dirlen,
+                                        child_l->table_name_length);
+      dir_path[dirlen - 1]= 0;
+      dirlen= dirname_length(dir_path);
+      child_l->db_length= strlen(dir_path + dirlen);
+      child_l->db= strmake_root(&parent->mem_root, dir_path + dirlen,
+                                child_l->db_length);
+    }
+  }
+
+  DBUG_PRINT("myrg", ("open: '%.*s'.'%.*s'", (int) child_l->db_length,
+                      child_l->db, (int) child_l->table_name_length,
+                      child_l->table_name));
+
   /* Convert to lowercase if required. */
   if (lower_case_table_names && child_l->table_name_length)
     child_l->table_name_length= my_casedn_str(files_charset_info,
@@ -306,7 +343,7 @@ static MI_INFO *myisammrg_attach_children_callback(void *callback_param)
   TABLE         *parent;
   TABLE         *child;
   TABLE_LIST    *child_l;
-  MI_INFO       *myisam;
+  MI_INFO       *UNINIT_VAR(myisam);
   DBUG_ENTER("myisammrg_attach_children_callback");
 
   my_errno= 0;
@@ -379,7 +416,7 @@ static MI_INFO *myisammrg_attach_children_callback(void *callback_param)
     my_errno= HA_ERR_WRONG_MRG_TABLE_DEF;
   }
   DBUG_PRINT("myrg", ("MyISAM handle: 0x%lx  my_errno: %d",
-                      (long) myisam, my_errno));
+                      my_errno ? 0L : (long) myisam, my_errno));
 
  err:
   DBUG_RETURN(my_errno ? NULL : myisam);
@@ -392,28 +429,49 @@ static MI_INFO *myisammrg_attach_children_callback(void *callback_param)
   @detail This function initializes the MERGE storage engine structures
     and adds a child list of TABLE_LIST to the parent TABLE.
 
-  @param[in]    name            MERGE table path name
-  @param[in]    mode            read/write mode, unused
-  @param[in]    test_if_locked  open flags
+  @param[in]    name                MERGE table path name
+  @param[in]    mode                read/write mode, unused
+  @param[in]    test_if_locked_arg  open flags
 
   @return       status
-    @retval     0               OK
-    @retval     -1              Error, my_errno gives reason
+    @retval     0                   OK
+    @retval     -1                  Error, my_errno gives reason
 */
 
 int ha_myisammrg::open(const char *name, int mode __attribute__((unused)),
-                       uint test_if_locked)
+                       uint test_if_locked_arg)
 {
   DBUG_ENTER("ha_myisammrg::open");
   DBUG_PRINT("myrg", ("name: '%s'  table: 0x%lx", name, (long) table));
-  DBUG_PRINT("myrg", ("test_if_locked: %u", test_if_locked));
+  DBUG_PRINT("myrg", ("test_if_locked_arg: %u", test_if_locked_arg));
 
   /* Save for later use. */
-  this->test_if_locked= test_if_locked;
+  test_if_locked= test_if_locked_arg;
 
   /* retrieve children table list. */
   my_errno= 0;
-  if (!(file= myrg_parent_open(name, myisammrg_parent_open_callback, this)))
+  if (is_cloned)
+  {
+    /*
+      Open and attaches the MyISAM tables,that are under the MERGE table 
+      parent, on the MyISAM storage engine interface directly within the
+      MERGE engine. The new MyISAM table instances, as well as the MERGE 
+      clone itself, are not visible in the table cache. This is not a 
+      problem because all locking is handled by the original MERGE table
+      from which this is cloned of.
+    */
+    if (!(file= myrg_open(table->s->normalized_path.str, table->db_stat, 
+                                       HA_OPEN_IGNORE_IF_LOCKED)))
+    {
+      DBUG_PRINT("error", ("my_errno %d", my_errno));
+      DBUG_RETURN(my_errno ? my_errno : -1); 
+    }
+
+    file->children_attached= TRUE;
+
+    info(HA_STATUS_NO_LOCK | HA_STATUS_VARIABLE | HA_STATUS_CONST);
+  }
+  else if (!(file= myrg_parent_open(name, myisammrg_parent_open_callback, this)))
   {
     DBUG_PRINT("error", ("my_errno %d", my_errno));
     DBUG_RETURN(my_errno ? my_errno : -1);
@@ -422,6 +480,55 @@ int ha_myisammrg::open(const char *name, int mode __attribute__((unused)),
   DBUG_RETURN(0);
 }
 
+/**
+   Returns a cloned instance of the current handler.
+
+   @return A cloned handler instance.
+ */
+handler *ha_myisammrg::clone(MEM_ROOT *mem_root)
+{
+  MYRG_TABLE    *u_table,*newu_table;
+  ha_myisammrg *new_handler= 
+    (ha_myisammrg*) get_new_handler(table->s, mem_root, table->s->db_type());
+  if (!new_handler)
+    return NULL;
+  
+  /* Inform ha_myisammrg::open() that it is a cloned handler */
+  new_handler->is_cloned= TRUE;
+  /*
+    Allocate handler->ref here because otherwise ha_open will allocate it
+    on this->table->mem_root and we will not be able to reclaim that memory 
+    when the clone handler object is destroyed.
+  */
+  if (!(new_handler->ref= (uchar*) alloc_root(mem_root, ALIGN_SIZE(ref_length)*2)))
+  {
+    delete new_handler;
+    return NULL;
+  }
+
+  if (new_handler->ha_open(table, table->s->normalized_path.str, table->db_stat,
+                            HA_OPEN_IGNORE_IF_LOCKED))
+  {
+    delete new_handler;
+    return NULL;
+  }
+ 
+  /*
+    Iterate through the original child tables and
+    copy the state into the cloned child tables.
+    We need to do this because all the child tables
+    can be involved in delete.
+  */
+  newu_table= new_handler->file->open_tables;
+  for (u_table= file->open_tables; u_table < file->end_table; u_table++)
+  {
+    newu_table->table->state= u_table->table->state;
+    newu_table++;
+  }
+
+  return new_handler;
+ }
+
 
 /**
   @brief Attach children to a MERGE table.
@@ -474,7 +581,8 @@ int ha_myisammrg::attach_children(void)
 
   if (myrg_attach_children(this->file, this->test_if_locked |
                            current_thd->open_options,
-                           myisammrg_attach_children_callback, this))
+                           myisammrg_attach_children_callback, this,
+                           (my_bool *) &need_compat_check))
   {
     DBUG_PRINT("error", ("my_errno %d", my_errno));
     DBUG_RETURN(my_errno ? my_errno : -1);
@@ -525,7 +633,7 @@ int ha_myisammrg::attach_children(void)
       if (check_definition(keyinfo, recinfo, keys, recs,
                            u_table->table->s->keyinfo, u_table->table->s->rec,
                            u_table->table->s->base.keys,
-                           u_table->table->s->base.fields, false))
+                           u_table->table->s->base.fields, false, NULL))
       {
         DBUG_PRINT("error", ("table definition mismatch: '%s'",
                              u_table->table->filename));
@@ -613,9 +721,10 @@ int ha_myisammrg::close(void)
   DBUG_ENTER("ha_myisammrg::close");
   /*
     Children must not be attached here. Unless the MERGE table has no
-    children. In this case children_attached is always true.
+    children or the handler instance has been cloned. In these cases 
+    children_attached is always true. 
   */
-  DBUG_ASSERT(!this->file->children_attached || !this->file->tables);
+  DBUG_ASSERT(!this->file->children_attached || !this->file->tables || this->is_cloned);
   rc= myrg_close(file);
   file= 0;
   DBUG_RETURN(rc);
@@ -801,7 +910,16 @@ int ha_myisammrg::info(uint flag)
     table->s->crashed= 1;
 #endif
   stats.data_file_length= mrg_info.data_file_length;
-  errkey= mrg_info.errkey;
+  if (mrg_info.errkey >= (int) table_share->keys)
+  {
+    /*
+     If value of errkey is higher than the number of keys
+     on the table set errkey to MAX_KEY. This will be
+     treated as unknown key case and error message generator
+     won't try to locate key causing segmentation fault.
+    */
+    mrg_info.errkey= MAX_KEY;
+  }
   table->s->keys_in_use.set_prefix(table->s->keys);
   stats.mean_rec_length= mrg_info.reclength;
   
@@ -843,14 +961,19 @@ int ha_myisammrg::info(uint flag)
         with such a number, it'll be an error later anyway.
       */
       bzero((char*) table->key_info[0].rec_per_key,
-            sizeof(table->key_info[0].rec_per_key) * table->s->key_parts);
+            sizeof(table->key_info[0].rec_per_key[0]) * table->s->key_parts);
 #endif
       memcpy((char*) table->key_info[0].rec_per_key,
 	     (char*) mrg_info.rec_per_key,
-             sizeof(table->key_info[0].rec_per_key) *
+             sizeof(table->key_info[0].rec_per_key[0]) *
              min(file->keys, table->s->key_parts));
     }
   }
+  if (flag & HA_STATUS_ERRKEY)
+  {
+    errkey= mrg_info.errkey;
+    my_store_ptr(dup_ref, ref_length, mrg_info.dupp_key_pos);
+  }
   return 0;
 }
 
@@ -877,7 +1000,7 @@ int ha_myisammrg::extra(enum ha_extra_function operation)
   /* As this is just a mapping, we don't have to force the underlying
      tables to be closed */
   if (operation == HA_EXTRA_FORCE_REOPEN ||
-      operation == HA_EXTRA_PREPARE_FOR_DELETE)
+      operation == HA_EXTRA_PREPARE_FOR_DROP)
     return 0;
   return myrg_extra(file,operation,0);
 }
@@ -961,7 +1084,7 @@ THR_LOCK_DATA **ha_myisammrg::store_lock(THD *thd,
 static void split_file_name(const char *file_name,
 			    LEX_STRING *db, LEX_STRING *name)
 {
-  uint dir_length, prefix_length;
+  size_t dir_length, prefix_length;
   char buff[FN_REFLEN];
 
   db->length= 0;
@@ -1009,8 +1132,8 @@ void ha_myisammrg::update_create_info(HA_CREATE_INFO *create_info)
 	goto err;
 
       create_info->merge_list.elements++;
-      (*create_info->merge_list.next) = (uchar*) ptr;
-      create_info->merge_list.next= (uchar**) &ptr->next_local;
+      (*create_info->merge_list.next) = ptr;
+      create_info->merge_list.next= &ptr->next_local;
     }
     *create_info->merge_list.next=0;
   }
@@ -1032,9 +1155,9 @@ int ha_myisammrg::create(const char *name, register TABLE *form,
 {
   char buff[FN_REFLEN];
   const char **table_names, **pos;
-  TABLE_LIST *tables= (TABLE_LIST*) create_info->merge_list.first;
+  TABLE_LIST *tables= create_info->merge_list.first;
   THD *thd= current_thd;
-  uint dirlgt= dirname_length(name);
+  size_t dirlgt= dirname_length(name);
   DBUG_ENTER("ha_myisammrg::create");
 
   /* Allocate a table_names array in thread mem_root. */
@@ -1045,7 +1168,7 @@ int ha_myisammrg::create(const char *name, register TABLE *form,
   /* Create child path names. */
   for (pos= table_names; tables; tables= tables->next_local)
   {
-    const char *table_name;
+    const char *table_name= buff;
 
     /*
       Construct the path to the MyISAM table. Try to meet two conditions:
@@ -1071,10 +1194,12 @@ int ha_myisammrg::create(const char *name, register TABLE *form,
       as the MyISAM tables are from the same database as the MERGE table.
     */
     if ((dirname_length(buff) == dirlgt) && ! memcmp(buff, name, dirlgt))
-      table_name= tables->table_name;
-    else
-      if (! (table_name= thd->strmake(buff, length)))
-        DBUG_RETURN(HA_ERR_OUT_OF_MEM); /* purecov: inspected */
+    {
+      table_name+= dirlgt;
+      length-= dirlgt;
+    }
+    if (!(table_name= thd->strmake(table_name, length)))
+      DBUG_RETURN(HA_ERR_OUT_OF_MEM); /* purecov: inspected */
 
     *pos++= table_name;
   }
@@ -1093,28 +1218,31 @@ int ha_myisammrg::create(const char *name, register TABLE *form,
 void ha_myisammrg::append_create_info(String *packet)
 {
   const char *current_db;
-  uint db_length;
+  size_t db_length;
   THD *thd= current_thd;
-  MYRG_TABLE *open_table, *first;
+  TABLE_LIST *open_table, *first;
 
   if (file->merge_insert_method != MERGE_INSERT_DISABLED)
   {
     packet->append(STRING_WITH_LEN(" INSERT_METHOD="));
     packet->append(get_type(&merge_insert_method,file->merge_insert_method-1));
   }
+  /*
+    There is no sence adding UNION clause in case there is no underlying
+    tables specified.
+  */
+  if (file->open_tables == file->end_table)
+    return;
   packet->append(STRING_WITH_LEN(" UNION=("));
 
   current_db= table->s->db.str;
   db_length=  table->s->db.length;
 
-  for (first=open_table=file->open_tables ;
-       open_table != file->end_table ;
-       open_table++)
+  for (first= open_table= table->child_l;;
+       open_table= open_table->next_global)
   {
-    LEX_STRING db, name;
-    LINT_INIT(db.str);
+    LEX_STRING db= { open_table->db, open_table->db_length };
 
-    split_file_name(open_table->table->filename, &db, &name);
     if (open_table != first)
       packet->append(',');
     /* Report database for mapped table if it isn't in current database */
@@ -1125,7 +1253,10 @@ void ha_myisammrg::append_create_info(String *packet)
       append_identifier(thd, packet, db.str, db.length);
       packet->append('.');
     }
-    append_identifier(thd, packet, name.str, name.length);
+    append_identifier(thd, packet, open_table->table_name,
+                      open_table->table_name_length);
+    if (&open_table->next_global == table->child_last_l)
+      break;
   }
   packet->append(')');
 }
@@ -1148,6 +1279,12 @@ int ha_myisammrg::check(THD* thd, HA_CHECK_OPT* check_opt)
 }
 
 
+ha_rows ha_myisammrg::records()
+{
+  return myrg_records(file);
+}
+
+
 extern int myrg_panic(enum ha_panic_function flag);
 int myisammrg_panic(handlerton *hton, ha_panic_function flag)
 {
