@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #ifndef GLOBAL_SIGNAL_NUMBERS_H
 #define GLOBAL_SIGNAL_NUMBERS_H
@@ -142,8 +144,8 @@ extern const GlobalSignalNumber NO_OF_SIGNAL_NAMES;
 #define GSN_ACC_OVER_REC                81
 
 /* 83 unused */
-#define GSN_ACC_SCAN_INFO               84
-#define GSN_ACC_SCAN_INFO24             85
+#define GSN_ACC_SCAN_INFO               84 /* local */
+#define GSN_ALLOC_MEM_REQ               85 /* local */
 #define GSN_ACC_SCANCONF                86
 #define GSN_ACC_SCANREF                 87
 #define GSN_ACC_SCANREQ                 88
@@ -185,7 +187,7 @@ extern const GlobalSignalNumber NO_OF_SIGNAL_NAMES;
 #define GSN_ROUTE_ORD                   121
 #define GSN_NODE_VERSION_REP            122
 /* 123 unused */
-/* 124 unused */
+#define GSN_FSSUSPENDORD                124 /* local */
 #define GSN_CHECK_LCP_STOP              125
 #define GSN_CLOSE_COMCONF               126 /* local */
 #define GSN_CLOSE_COMREQ                127 /* local */
@@ -342,6 +344,7 @@ extern const GlobalSignalNumber NO_OF_SIGNAL_NAMES;
 #define GSN_END_LCP_REQ                 244
 #define GSN_END_TOCONF                  245
 #define GSN_END_TOREQ                   246
+#define GSN_END_TOREF                   286
 #define GSN_EVENT_REP                   247
 #define GSN_EXEC_FRAGCONF               248
 #define GSN_EXEC_FRAGREF                249
@@ -383,7 +386,7 @@ extern const GlobalSignalNumber NO_OF_SIGNAL_NAMES;
 
 #define GSN_UPGRADE_PROTOCOL_ORD        285
 
-/* 286 unused */
+/* 286 not unused */
 /* 287 unused */
 #define GSN_GETGCICONF                  288
 #define GSN_GETGCIREQ                   289
@@ -403,7 +406,7 @@ extern const GlobalSignalNumber NO_OF_SIGNAL_NAMES;
 #define GSN_GET_SCHEMA_INFOREQ          302
 /* 303 not unused */
 /* 304 not unused */
-#define GSN_LQH_RESTART_OP              305
+#define GSN_ALLOC_MEM_REF               305 /* local */
 #define GSN_LQH_TRANSCONF               306
 #define GSN_LQH_TRANSREQ                307
 #define GSN_LQHADDATTCONF               308
@@ -432,7 +435,7 @@ extern const GlobalSignalNumber NO_OF_SIGNAL_NAMES;
 #define GSN_NEXT_SCANCONF               330
 #define GSN_NEXT_SCANREF                331
 #define GSN_NEXT_SCANREQ                332
-#define GSN_NEXTOPERATION               333
+#define GSN_ALLOC_MEM_CONF              333 /* local */
 
 #define GSN_READ_CONFIG_REQ             334 /* new name for sizealt, local */
 #define GSN_READ_CONFIG_CONF            335 /* new name for sizealt, local */
@@ -494,11 +497,15 @@ extern const GlobalSignalNumber NO_OF_SIGNAL_NAMES;
 #define GSN_START_PERMCONF              381
 #define GSN_START_PERMREF               382
 #define GSN_START_PERMREQ               383
+#define GSN_START_PERMREP               422
 #define GSN_START_RECCONF               384
 #define GSN_START_RECREF                385
 #define GSN_START_RECREQ                386
+
 #define GSN_START_TOCONF                387
 #define GSN_START_TOREQ                 388
+#define GSN_START_TOREF                 421
+
 #define GSN_STORED_PROCCONF             389
 #define GSN_STORED_PROCREF              390
 #define GSN_STORED_PROCREQ              391
@@ -527,8 +534,8 @@ extern const GlobalSignalNumber NO_OF_SIGNAL_NAMES;
 #define GSN_TUP_ATTRINFO                418
 #define GSN_TUP_COMMITREQ               419
 
-/* 421 unused */
-/* 422 unused */
+/* 421 not unused */
+/* 422 not unused */
 /* 423 unused */
 
 /* 424 unused */
@@ -536,7 +543,7 @@ extern const GlobalSignalNumber NO_OF_SIGNAL_NAMES;
 /* 426 unused */
 /* 427 unused */
 /* 428 unused */
-/* 429 unused */
+#define GSN_LOCAL_ROUTE_ORD             429 /* local */
 /* 430 unused */
 #define GSN_TUPFRAGCONF                 431
 #define GSN_TUPFRAGREF                  432
