@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #ifndef NDB_BITMASK_H
 #define NDB_BITMASK_H
@@ -47,9 +49,10 @@ public:
   static void set(unsigned size, Uint32 data[]);
 
   /**
-   * set bit from <em>start</em> to <em>last</em>
+   * set <em>len</em> bist from <em>start</em>
    */
-  static void set_range(unsigned size, Uint32 data[], unsigned start, unsigned last);
+  static void setRange(unsigned size, Uint32 data[], unsigned start,
+                       unsigned len);
 
   /**
    * assign - Set all bits in <em>dst</em> to corresponding in <em>src/<em>
@@ -121,6 +124,11 @@ public:
    */
   static void bitXORC(unsigned size, Uint32 data[], const Uint32 data2[]);
 
+  /**
+   * bitNOT - Bitwise (~x) into first operand.
+   */
+  static void bitNOT(unsigned size, Uint32 data[]);
+
   /**
    * contains - Check if all bits set in data2 are set in data
    */
@@ -158,10 +166,28 @@ public:
   static void setField(unsigned size, Uint32 data[],
 		       unsigned pos, unsigned len, const Uint32 src[]);
   
+  /**
+   * copyField - Copy bitfield from one position and length
+   * to another position and length.
+   * Undefined for overlapping bitfields
+   */
+  static void copyField(Uint32 dst[], unsigned destPos,
+                        const Uint32 src[], unsigned srcPos, unsigned len);
+  
   /**
    * getText - Return as hex-digits (only for debug routines).
    */
   static char* getText(unsigned size, const Uint32 data[], char* buf);
+
+  /**
+   * Parse string with numbers format
+   *   1,2,3-5
+   * @return -1 if unparsable chars found, 
+   *         -2 str has number > bitmask size
+   *            else returns number of bits set 
+   */
+  static int parseMask(unsigned size, Uint32 data[], const char * str);
+
 private:
   static void getFieldImpl(const Uint32 data[], unsigned, unsigned, Uint32 []);
   static void setFieldImpl(Uint32 data[], unsigned, unsigned, const Uint32 []);
@@ -196,9 +222,10 @@ BitmaskImpl::set(unsigned size, Uint32 data[])
 }
 
 inline void
-BitmaskImpl::set_range(unsigned size, Uint32 data[], 
-		       unsigned start, unsigned last)
+BitmaskImpl::setRange(unsigned size, Uint32 data[],
+                      unsigned start, unsigned len)
 {
+  Uint32 last = start + len - 1;
   Uint32 *ptr = data + (start >> 5);
   Uint32 *end =  data + (last >> 5);
   assert(start <= last);
@@ -375,6 +402,14 @@ BitmaskImpl::bitXORC(unsigned size, Uint32 data[], const Uint32 data2[])
   }
 }
 
+inline void
+BitmaskImpl::bitNOT(unsigned size, Uint32 data[])
+{
+  for (unsigned i = 0; i < size; i++) {
+    data[i] = ~data[i];
+  }
+}
+
 inline bool
 BitmaskImpl::contains(unsigned size, Uint32 data[], const Uint32 data2[])
 {
@@ -502,6 +537,12 @@ public:
   static void set(Uint32 data[]);
   void set();
 
+  /**
+   * set - set a range of bits
+   */
+  static void setRange(Uint32 data[], Uint32 pos, Uint32 len);
+  void setRange(Uint32 pos, Uint32 len);
+
   /**
    * clear - Clear bit n.
    */
@@ -576,6 +617,12 @@ public:
   static void bitXORC(Uint32 data[], const Uint32 data2[]);
   BitmaskPOD<size>& bitXORC(const BitmaskPOD<size>& mask2);
 
+  /**
+   * bitNOT - Bitwise (~x) in first operand.
+   */
+  static void bitNOT(Uint32 data[]);
+  BitmaskPOD<size>& bitNOT();
+
   /**
    * contains - Check if all bits set in data2 (that) are also set in data (this)
    */
@@ -593,6 +640,9 @@ public:
    */
   static char* getText(const Uint32 data[], char* buf);
   char* getText(char* buf) const;
+
+  static int parseMask(Uint32 data[], const char * src);
+  int parseMask(const char * src);
 };
 
 template <unsigned size>
@@ -693,6 +743,21 @@ BitmaskPOD<size>::set()
   BitmaskPOD<size>::set(rep.data);
 }
 
+template <unsigned size>
+inline void
+BitmaskPOD<size>::setRange(Uint32 data[], Uint32 pos, Uint32 len)
+{
+  BitmaskImpl::setRange(size, data, pos, len);
+}
+
+
+template <unsigned size>
+inline void
+BitmaskPOD<size>::setRange(Uint32 pos, Uint32 len)
+{
+  BitmaskPOD<size>::setRange(rep.data, pos, len);
+}
+
 template <unsigned size>
 inline void
 BitmaskPOD<size>::clear(Uint32 data[], unsigned n)
@@ -750,7 +815,7 @@ BitmaskPOD<size>::isclear() const
 }
 
 template <unsigned size>
-unsigned
+inline unsigned
 BitmaskPOD<size>::count(const Uint32 data[])
 {
   return BitmaskImpl::count(size, data);
@@ -764,7 +829,7 @@ BitmaskPOD<size>::count() const
 }
 
 template <unsigned size>
-unsigned
+inline unsigned
 BitmaskPOD<size>::find(const Uint32 data[], unsigned n)
 {
   return BitmaskImpl::find(size, data, n);
@@ -867,7 +932,22 @@ BitmaskPOD<size>::bitXORC(const BitmaskPOD<size>& mask2)
 }
 
 template <unsigned size>
-char *
+inline void
+BitmaskPOD<size>::bitNOT(Uint32 data[])
+{
+  BitmaskImpl::bitNOT(size,data);
+}
+
+template <unsigned size>
+inline BitmaskPOD<size>&
+BitmaskPOD<size>::bitNOT()
+{
+  BitmaskPOD<size>::bitNOT(rep.data);
+  return *this;
+}
+
+template <unsigned size>
+inline char *
 BitmaskPOD<size>::getText(const Uint32 data[], char* buf)
 {
   return BitmaskImpl::getText(size, data, buf);
@@ -908,6 +988,21 @@ BitmaskPOD<size>::overlaps(BitmaskPOD<size> that)
   return BitmaskPOD<size>::overlaps(this->rep.data, that.rep.data);
 }
 
+template <unsigned size>
+int
+BitmaskPOD<size>::parseMask(Uint32 data[], const char* buf)
+{
+  return BitmaskImpl::parseMask(size, data, buf);
+}
+
+template <unsigned size>
+inline
+int
+BitmaskPOD<size>::parseMask(const char* buf)
+{
+  return BitmaskPOD<size>::parseMask(rep.data, buf);
+}
+
 template <unsigned size>
 class Bitmask : public BitmaskPOD<size> {
 public:
@@ -979,4 +1074,58 @@ BitmaskImpl::setField(unsigned size, Uint32 dst[],
   setFieldImpl(dst+1, used & 31, len-used, src+(used >> 5));
 }
 
+/* Three way min utiltiy for copyField below */
+inline unsigned minLength(unsigned a, unsigned b, unsigned c)
+{
+  return (a < b ? 
+          (a < c ? a : c) : 
+          (b < c ? b : c ));
+}
+
+inline void
+BitmaskImpl::copyField(Uint32 _dst[], unsigned dstPos,
+                       const Uint32 _src[], unsigned srcPos, unsigned len)
+{
+  /* Algorithm
+   * While (len > 0)
+   *  - Find the longest bit length we can copy from one 32-bit word
+   *    to another (which is the miniumum of remaining length, 
+   *    space in current src word and space in current dest word)
+   *  - Extract that many bits from src, and shift them to the correct
+   *    position to insert into dest
+   *  - Mask out the to-be-written words from dest (and any irrelevant 
+   *    words in src) and or them together
+   *  - Move onto next chunk
+   */
+  while (len > 0)
+  {
+    const Uint32* src= _src + (srcPos >> 5);
+    Uint32* dst= _dst + (dstPos >> 5);
+    unsigned srcOffset= srcPos & 31;
+    unsigned dstOffset= dstPos & 31;
+    unsigned srcBitsInWord= 32 - srcOffset; 
+    unsigned dstBitsInWord= 32 - dstOffset;
+    
+    /* How many bits can we copy at once? */
+    unsigned bits= minLength(dstBitsInWord, srcBitsInWord, len);
+    
+    /* Create mask for affected bits in dest */
+    Uint32 destMask= (~(Uint32)0 >> (32-bits) << dstOffset);
+    
+    /* Grab source data and shift to dest offset */
+    Uint32 data= ((*src) >> srcOffset) << dstOffset;
+    
+    /* Mask out affected bits in dest and irrelevant bits in source
+     * and combine
+     */
+    *dst= (*dst  & ~destMask) | (data & destMask);
+    
+    srcPos+= bits;
+    dstPos+= bits;
+    len-= bits;
+  }
+  
+  return;
+}
+
 #endif
