@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #define DBTUP_C
 #define DBTUP_COMMIT_CPP
@@ -116,13 +118,39 @@ void Dbtup::initOpConnection(Operationrec* regOperPtr)
   regOperPtr->m_undo_buffer_space= 0;
 }
 
-static
-inline
 bool
-operator>(const Local_key& key1, const Local_key& key2)
+Dbtup::is_rowid_lcp_scanned(const Local_key& key1,
+                            const Dbtup::ScanOp& op)
 {
-  return key1.m_page_no > key2.m_page_no ||
-    (key1.m_page_no == key2.m_page_no && key1.m_page_idx > key2.m_page_idx);
+  Local_key key2 = op.m_scanPos.m_key;
+  switch (op.m_state) {
+  case Dbtup::ScanOp::First:
+    ndbrequire(key2.isNull());
+    return false;
+  case Dbtup::ScanOp::Current:
+  case Dbtup::ScanOp::Next:
+    ndbrequire(!key2.isNull());
+    if (key1.m_page_no < key2.m_page_no)
+      return true;
+    if (key1.m_page_no > key2.m_page_no)
+      return false;
+    if (key1.m_page_idx < key2.m_page_idx)
+      return true;
+    if (key1.m_page_idx > key2.m_page_idx)
+      return false;
+    /**
+     * key are equal...need to look at scan state
+     */
+    if (op.m_state == Dbtup::ScanOp::Next)
+      return true;
+    return false;
+  case Dbtup::ScanOp::Last:
+    return true;
+  default:
+    ndbrequire(false);
+    break;
+  }
+  return false;
 }
 
 void
@@ -152,15 +180,14 @@ Dbtup::dealloc_tuple(Signal* signal,
   }
   
   if (! (bits & (Tuple_header::LCP_SKIP | Tuple_header::ALLOC)) && 
-      lcpScan_ptr_i != RNIL)
+      lcpScan_ptr_i != RNIL && regTabPtr->m_no_of_disk_attributes > 0)
   {
     jam();
     ScanOpPtr scanOp;
     c_scanOpPool.getPtr(scanOp, lcpScan_ptr_i);
     Local_key rowid = regOperPtr->m_tuple_location;
-    Local_key scanpos = scanOp.p->m_scanPos.m_key;
     rowid.m_page_no = page->frag_page_id;
-    if (rowid > scanpos)
+    if (!is_rowid_lcp_scanned(rowid, *scanOp.p))
     {
       jam();
       extra_bits = Tuple_header::LCP_KEEP; // Note REMOVE FREE
@@ -178,6 +205,7 @@ Dbtup::dealloc_tuple(Signal* signal,
   }
 }
 
+#if 0
 static void dump_buf_hex(unsigned char *p, Uint32 bytes)
 {
   char buf[3001];
@@ -195,6 +223,8 @@ static void dump_buf_hex(unsigned char *p, Uint32 bytes)
   }
   ndbout_c("%8p: %s", p, buf);
 }
+#endif
+
 void
 Dbtup::commit_operation(Signal* signal,
 			Uint32 gci,
@@ -211,8 +241,7 @@ Dbtup::commit_operation(Signal* signal,
   Uint32 bits= tuple_ptr->m_header_bits;
 
   Tuple_header *disk_ptr= 0;
-  Tuple_header *copy= (Tuple_header*)
-    c_undo_buffer.get_ptr(&regOperPtr->m_copy_tuple_location);
+  Tuple_header *copy= get_copy_tuple(&regOperPtr->m_copy_tuple_location);
   
   Uint32 copy_bits= copy->m_header_bits;
 
@@ -244,12 +273,13 @@ Dbtup::commit_operation(Signal* signal,
     if (copy_bits & Tuple_header::VAR_PART)
     {
       jam();
-      ndbassert(tmp.m_page_no != RNIL);
       ndbassert(bits & Tuple_header::VAR_PART);
+      ndbassert(tmp.m_page_no != RNIL);
+      ndbassert(copy_bits & Tuple_header::COPY_TUPLE);
+
       Uint32 *dst= get_ptr(&vpagePtr, *ref);
       Var_page* vpagePtrP = (Var_page*)vpagePtr.p;
       Varpart_copy*vp =(Varpart_copy*)copy->get_end_of_fix_part_ptr(regTabPtr);
-      ndbassert(copy_bits & Tuple_header::COPY_TUPLE);
       /* The first word of shrunken tuple holds the lenght in words. */
       Uint32 len = vp->m_len;
       memcpy(dst, vp->m_data, 4*len);
@@ -262,16 +292,16 @@ Dbtup::commit_operation(Signal* signal,
         {
           jam();
           vpagePtrP->shrink_entry(tmp.m_page_idx, len);
+          update_free_page_list(regFragPtr, vpagePtr);
         }
         else
         {
           jam();
-          vpagePtrP->free_record(tmp.m_page_idx, Var_page::CHAIN);
+          free_var_part(regFragPtr, vpagePtr, tmp.m_page_idx);
           tmp.m_page_no = RNIL;
           ref->assign(&tmp);
           copy_bits &= ~(Uint32)Tuple_header::VAR_PART;
         }
-        update_free_page_list(regFragPtr, vpagePtr);
       }
       else
       {
@@ -301,7 +331,7 @@ Dbtup::commit_operation(Signal* signal,
     memcpy(&key, copy->get_disk_ref_ptr(regTabPtr), sizeof(Local_key));
     Uint32 logfile_group_id= regFragPtr->m_logfile_group_id;
 
-    PagePtr diskPagePtr = *(PagePtr*)&m_pgman.m_ptr;
+    PagePtr diskPagePtr = { (Tup_page*)m_pgman.m_ptr.p, m_pgman.m_ptr.i };
     ndbassert(diskPagePtr.p->m_page_no == key.m_page_no);
     ndbassert(diskPagePtr.p->m_file_no == key.m_file_no);
     Uint32 sz, *dst;
@@ -344,9 +374,8 @@ Dbtup::commit_operation(Signal* signal,
     ScanOpPtr scanOp;
     c_scanOpPool.getPtr(scanOp, lcpScan_ptr_i);
     Local_key rowid = regOperPtr->m_tuple_location;
-    Local_key scanpos = scanOp.p->m_scanPos.m_key;
     rowid.m_page_no = pagePtr.p->frag_page_id;
-    if(rowid > scanpos)
+    if (!is_rowid_lcp_scanned(rowid, *scanOp.p))
     {
       jam();
        copy_bits |= Tuple_header::LCP_SKIP;
@@ -576,8 +605,7 @@ void Dbtup::execTUP_COMMITREQ(Signal* signal)
     if(!regOperPtr.p->m_copy_tuple_location.isNull())
     {
       jam();
-      Tuple_header* tmp= (Tuple_header*)
-	c_undo_buffer.get_ptr(&regOperPtr.p->m_copy_tuple_location);
+      Tuple_header* tmp= get_copy_tuple(&regOperPtr.p->m_copy_tuple_location);
       
       memcpy(&req.m_page, 
 	     tmp->get_disk_ref_ptr(regTabPtr.p), sizeof(Local_key));
@@ -704,7 +732,7 @@ skip_disk:
      * Perform "real" commit
      */
     Uint32 disk = regOperPtr.p->m_commit_disk_callback_page;
-    set_change_mask_info(&req_struct, regOperPtr.p);
+    set_commit_change_mask_info(regTabPtr.p, &req_struct, regOperPtr.p);
     checkDetachedTriggers(&req_struct, regOperPtr.p, regTabPtr.p, 
                           disk != RNIL);
     
@@ -720,7 +748,9 @@ skip_disk:
     {
       jam();
       if (get_page)
+      {
 	ndbassert(tuple_ptr->m_header_bits & Tuple_header::DISK_PART);
+      }
       dealloc_tuple(signal, gci_hi, page.p, tuple_ptr,
 		    regOperPtr.p, regFragPtr.p, regTabPtr.p); 
     }
@@ -747,58 +777,31 @@ skip_disk:
 }
 
 void
-Dbtup::set_change_mask_info(KeyReqStruct * const req_struct,
-                            Operationrec * const regOperPtr)
+Dbtup::set_commit_change_mask_info(const Tablerec* regTabPtr,
+                                   KeyReqStruct * req_struct,
+                                   const Operationrec * regOperPtr)
 {
-  ChangeMaskState state = get_change_mask_state(regOperPtr);
-  if (state == USE_SAVED_CHANGE_MASK) {
-    jam();
-    req_struct->changeMask.setWord(0, regOperPtr->saved_change_mask[0]);
-    req_struct->changeMask.setWord(1, regOperPtr->saved_change_mask[1]);
-  } else if (state == RECALCULATE_CHANGE_MASK) {
-    jam();
-    // Recompute change mask, for now set all bits
-    req_struct->changeMask.set();
-  } else if (state == SET_ALL_MASK) {
-    jam();
-    req_struct->changeMask.set();
-  } else {
-    jam();
-    ndbrequire(state == DELETE_CHANGES);
+  Uint32 masklen = (regTabPtr->m_no_of_attributes + 31) >> 5;
+  if (regOperPtr->m_copy_tuple_location.isNull())
+  {
+    ndbassert(regOperPtr->op_struct.op_type == ZDELETE);
     req_struct->changeMask.set();
   }
-}
-
-void
-Dbtup::calculateChangeMask(Page* const pagePtr,
-                           Tablerec* const regTabPtr,
-                           KeyReqStruct * const req_struct)
-{
-  OperationrecPtr loopOpPtr;
-  Uint32 saved_word1= 0;
-  Uint32 saved_word2= 0;
-  loopOpPtr.i= req_struct->m_tuple_ptr->m_operation_ptr_i;
-  do {
-    c_operation_pool.getPtr(loopOpPtr);
-    ndbrequire(loopOpPtr.p->op_struct.op_type == ZUPDATE);
-    ChangeMaskState change_mask= get_change_mask_state(loopOpPtr.p);
-    if (change_mask == USE_SAVED_CHANGE_MASK) {
-      jam();
-      saved_word1|= loopOpPtr.p->saved_change_mask[0];
-      saved_word2|= loopOpPtr.p->saved_change_mask[1];
-    } else if (change_mask == RECALCULATE_CHANGE_MASK) {
-      jam();
-      //Recompute change mask, for now set all bits
-      req_struct->changeMask.set();
-      return;
-    } else {
-      ndbrequire(change_mask == SET_ALL_MASK);
-      jam();
-      req_struct->changeMask.set();
-      return;
+  else
+  {
+    Uint32 * dst = req_struct->changeMask.rep.data;
+    Uint32 * maskptr = get_copy_tuple_raw(&regOperPtr->m_copy_tuple_location);
+    Uint32 cols = * maskptr;
+    if (cols == regTabPtr->m_no_of_attributes)
+    {
+      memcpy(dst, maskptr + 1, 4*masklen);
     }
-    loopOpPtr.i= loopOpPtr.p->prevActiveOp;
-  } while (loopOpPtr.i != RNIL);
-  req_struct->changeMask.setWord(0, saved_word1);
-  req_struct->changeMask.setWord(1, saved_word2);
+    else
+    {
+      ndbassert(regTabPtr->m_no_of_attributes > cols); // no drop column
+      memcpy(dst, maskptr + 1, 4*((cols + 31) >> 5));
+      req_struct->changeMask.setRange(cols,
+                                      regTabPtr->m_no_of_attributes - cols);
+    }
+  }
 }
