@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2005, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include "pgman.hpp"
 #include <signaldata/FsRef.hpp>
@@ -31,14 +33,6 @@
                      Page_request::DIRTY_REQ | \
                      Page_request::ALLOC_REQ)
 
-// todo use this
-#ifdef VM_TRACE
-#define dbg(x) \
-  do { if (! debugFlag) break; debugOut << "PGMAN: " << x << endl; } while (0)
-#else
-#define dbg(x)
-#endif
-
 static bool g_dbg_lcp = false;
 #if 1
 #define DBG_LCP(x)
@@ -55,6 +49,7 @@ Pgman::Pgman(Block_context& ctx) :
 #ifdef VM_TRACE
   ,debugOut(* new NullOutputStream())
   ,debugFlag(false)
+  ,debugSummaryFlag(false)
 #endif
 {
   BLOCK_CONSTRUCTOR(Pgman);
@@ -76,9 +71,9 @@ Pgman::Pgman(Block_context& ctx) :
   m_stats_loop_on = false;
   m_busy_loop_on = false;
   m_cleanup_loop_on = false;
-  m_lcp_loop_on = false;
 
   // LCP variables
+  m_lcp_state = LS_LCP_OFF;
   m_last_lcp = 0;
   m_last_lcp_complete = 0;
   m_lcp_curr_bucket = ~(Uint32)0;
@@ -122,10 +117,18 @@ Pgman::execREAD_CONFIG_REQ(Signal* signal)
   
   if (page_buffer > 0)
   {
-    page_buffer = (page_buffer + GLOBAL_PAGE_SIZE - 1) / GLOBAL_PAGE_SIZE; // in pages
-    m_param.m_max_pages = page_buffer;
-    m_page_entry_pool.setSize(m_param.m_lirs_stack_mult * page_buffer);
-    m_param.m_max_hot_pages = (page_buffer * 9) / 10;
+    Uint32 page_cnt = Uint32((page_buffer + GLOBAL_PAGE_SIZE - 1) / GLOBAL_PAGE_SIZE);
+
+    if (ERROR_INSERTED(11009))
+    {
+      page_cnt = 25;
+      ndbout_c("Setting page_cnt = %u", page_cnt);
+    }
+
+    m_param.m_max_pages = page_cnt;
+    m_page_entry_pool.setSize(m_param.m_lirs_stack_mult * page_cnt);
+    m_param.m_max_hot_pages = (page_cnt * 9) / 10;
+    ndbrequire(m_param.m_max_hot_pages >= 1);
   }
 
   Pool_context pc;
@@ -153,9 +156,10 @@ Pgman::Param::Param() :
 
 Pgman::Stats::Stats() :
   m_num_pages(0),
+  m_num_hot_pages(0),
+  m_current_io_waits(0),
   m_page_hits(0),
-  m_page_faults(0),
-  m_current_io_waits(0)
+  m_page_faults(0)
 {
 }
 
@@ -233,11 +237,13 @@ Pgman::execCONTINUEB(Signal* signal)
     Page_sublist& pl = *m_page_sublist[Page_entry::SL_LOCKED];
     if (data1 != RNIL)
     {
+      jam();
       pl.getPtr(ptr, data1);
       process_lcp_locked(signal, ptr);
     }
     else
     {
+      jam();
       if (ERROR_INSERTED(11007))
       {
         ndbout << "No more writes..." << endl;
@@ -247,6 +253,7 @@ Pgman::execCONTINUEB(Signal* signal)
       }
       signal->theData[0] = m_end_lcp_req.senderData;
       sendSignal(m_end_lcp_req.senderRef, GSN_END_LCP_CONF, signal, 1, JBB);
+      m_lcp_state = LS_LCP_OFF;
     }
     return;
   }
@@ -259,14 +266,14 @@ Pgman::execCONTINUEB(Signal* signal)
 // page entry
 
 Pgman::Page_entry::Page_entry(Uint32 file_no, Uint32 page_no) :
-  m_state(0),
   m_file_no(file_no),
+  m_state(0),
   m_page_no(page_no),
   m_real_page_i(RNIL),
-  m_copy_page_i(RNIL),
   m_lsn(0),
   m_last_lcp(0),
   m_dirty_count(0),
+  m_copy_page_i(RNIL),
   m_busy_count(0),
   m_requests()
 {
@@ -277,10 +284,6 @@ Pgman::Page_entry::Page_entry(Uint32 file_no, Uint32 page_no) :
 Uint32
 Pgman::get_sublist_no(Page_state state)
 {
-  if (state == 0)
-  {
-    return ZNIL;
-  }
   if (state & Page_entry::REQUEST)
   {
     if (! (state & Page_entry::BOUND))
@@ -312,7 +315,11 @@ Pgman::get_sublist_no(Page_state state)
   if (state == Page_entry::ONSTACK) {
     return Page_entry::SL_IDLE;
   }
-  return Page_entry::SL_OTHER;
+  if (state != 0)
+  {
+    return Page_entry::SL_OTHER;
+  }
+  return ZNIL;
 }
 
 void
@@ -347,10 +354,25 @@ Pgman::set_page_state(Ptr<Page_entry> ptr, Page_state new_state)
       }
     }
     ptr.p->m_state = new_state;
+
+    bool old_hot = (old_state & Page_entry::HOT);
+    bool new_hot = (new_state & Page_entry::HOT);
+    if (! old_hot && new_hot)
+    {
+      jam();
+      m_stats.m_num_hot_pages++;
+    }
+    if (old_hot && ! new_hot)
+    {
+      jam();
+      ndbrequire(m_stats.m_num_hot_pages != 0);
+      m_stats.m_num_hot_pages--;
+    }
   }
 
 #ifdef VM_TRACE
   debugOut << "PGMAN: " << ptr << ": after" << endl;
+  verify_page_entry(ptr);
   debugOut << "PGMAN: <set_page_state" << endl;
 #endif
 }
@@ -499,7 +521,9 @@ Pgman::release_page_entry(Ptr<Page_entry>& ptr)
   ndbrequire(ptr.p->m_real_page_i == RNIL);
 
   if (! (state & Page_entry::LOCKED))
+  {
     ndbrequire(! (state & Page_entry::REQUEST));
+  }
 
   if (ptr.p->m_copy_page_i != RNIL)
   {
@@ -632,8 +656,10 @@ Pgman::lirs_reference(Ptr<Page_entry> ptr)
   Page_state state = ptr.p->m_state;
   ndbrequire(! (state & Page_entry::LOCKED));
 
-  // even non-LIRS cache pages are counted on l.h.s.
-  if (m_stats.m_num_pages >= m_param.m_max_hot_pages)
+  ndbrequire(m_stats.m_num_hot_pages <= m_param.m_max_hot_pages);
+
+  // LIRS kicks in when we have max hot pages
+  if (m_stats.m_num_hot_pages == m_param.m_max_hot_pages)
   {
     if (state & Page_entry::HOT)
     {
@@ -675,6 +701,12 @@ Pgman::lirs_reference(Ptr<Page_entry> ptr)
       jam();
       pl_stack.add(ptr);
       state |= Page_entry::ONSTACK;
+      /*
+       * bug#48910.  Using hot page count (not total page count)
+       * guarantees that stack is not empty here.  Therefore the new
+       * entry (added to top) is not at bottom and need not be hot.
+       */
+      ndbrequire(pl_stack.hasPrev(ptr));
       if (state & Page_entry::ONQUEUE)
       {
         jam();
@@ -698,8 +730,8 @@ Pgman::lirs_reference(Ptr<Page_entry> ptr)
   else
   {
 #ifdef VM_TRACE
-    debugOut << "PGMAN: filling up initial hot pages: "
-             << m_stats.m_num_pages << " of "
+    debugOut << "PGMAN: filling up hot pages: "
+             << m_stats.m_num_hot_pages << "/"
              << m_param.m_max_hot_pages << endl;
 #endif
     jam();
@@ -802,40 +834,35 @@ Pgman::do_cleanup_loop(Signal* signal)
 }
 
 void
-Pgman::do_lcp_loop(Signal* signal, bool direct)
+Pgman::do_lcp_loop(Signal* signal)
 {
 #ifdef VM_TRACE
-  debugOut << "PGMAN: >do_lcp_loop on=" << m_lcp_loop_on
-           << " direct=" << direct << endl;
+  debugOut << "PGMAN: >do_lcp_loop m_lcp_state=" << Uint32(m_lcp_state)
+           << endl;
 #endif
-  Uint32 restart = false;
-  if (direct)
-  {
-    ndbrequire(! m_lcp_loop_on);
-    restart = true;
-    m_lcp_loop_on = true;
-  }
-  else
-  {
-    ndbrequire(m_lcp_loop_on);
-    restart += process_lcp(signal);
-    if (! restart)
-    {
-      m_lcp_loop_on = false;
-    }
-  }
-  if (restart)
-  {
-    Uint32 delay = m_param.m_lcp_loop_delay;
+  ndbrequire(m_lcp_state != LS_LCP_OFF);
+  LCP_STATE newstate = process_lcp(signal);
+
+  switch(newstate) {
+  case LS_LCP_OFF:
+    jam();
+    break;
+  case LS_LCP_ON:
+    jam();
     signal->theData[0] = PgmanContinueB::LCP_LOOP;
-    if (delay)
-      sendSignalWithDelay(PGMAN_REF, GSN_CONTINUEB, signal, delay, 1);
-    else
-      sendSignal(PGMAN_REF, GSN_CONTINUEB, signal, 1, JBB);
+    sendSignal(PGMAN_REF, GSN_CONTINUEB, signal, 1, JBB);
+    break;
+  case LS_LCP_MAX_LCP_OUTSTANDING: // wait until io is completed
+    jam();
+    break;
+  case LS_LCP_LOCKED:
+    jam();
+    break;
   }
+  m_lcp_state = newstate;
 #ifdef VM_TRACE
-  debugOut << "PGMAN: <do_lcp_loop on=" << m_lcp_loop_on
-           << " restart=" << restart << endl;
+  debugOut << "PGMAN: <do_lcp_loop m_lcp_state=" << m_lcp_state
+           << endl;
 #endif
 }
 
@@ -1043,11 +1070,11 @@ Pgman::process_callback(Signal* signal, Ptr<Page_entry> ptr)
   debugOut << "PGMAN: " << ptr << " : process_callback" << endl;
 #endif
   int max_count = 1;
-  Page_state state = ptr.p->m_state;
 
   while (! ptr.p->m_requests.isEmpty() && --max_count >= 0)
   {
     jam();
+    Page_state state = ptr.p->m_state;
     SimulatedBlock* b;
     Callback callback;
     {
@@ -1089,19 +1116,18 @@ Pgman::process_callback(Signal* signal, Ptr<Page_entry> ptr)
     }
     ndbrequire(state & Page_entry::BOUND);
     ndbrequire(state & Page_entry::MAPPED);
+
+    // make REQUEST state consistent before set_page_state()
+    if (ptr.p->m_requests.isEmpty())
+    {
+      jam();
+      state &= ~ Page_entry::REQUEST;
+    }
     
     // callback may re-enter PGMAN and change page state
     set_page_state(ptr, state);
     b->execute(signal, callback, ptr.p->m_real_page_i);
-    state = ptr.p->m_state;
   }
-  
-  if (ptr.p->m_requests.isEmpty())
-  {
-    jam();
-    state &= ~ Page_entry::REQUEST;
-  }
-  set_page_state(ptr, state);
   return true;
 }
 
@@ -1237,17 +1263,20 @@ Pgman::execEND_LCP_REQ(Signal* signal)
 #endif
 
   m_last_lcp_complete = m_last_lcp;
-  
-  do_lcp_loop(signal, true);
+  ndbrequire(m_lcp_state == LS_LCP_OFF);
+  m_lcp_state = LS_LCP_ON;
+  do_lcp_loop(signal);
 }
 
-bool
+Pgman::LCP_STATE
 Pgman::process_lcp(Signal* signal)
 {
   Page_hashlist& pl_hash = m_page_hashlist;
 
   int max_count = 0;
-  if (m_param.m_max_io_waits > m_stats.m_current_io_waits) {
+  if (m_param.m_max_io_waits > m_stats.m_current_io_waits)
+  {
+    jam();
     max_count = m_param.m_max_io_waits - m_stats.m_current_io_waits;
     max_count = max_count / 2 + 1;
   }
@@ -1263,6 +1292,7 @@ Pgman::process_lcp(Signal* signal)
   // start or re-start from beginning of current hash bucket
   if (m_lcp_curr_bucket != ~(Uint32)0)
   {
+    jam();
     Page_hashlist::Iterator iter;
     pl_hash.next(m_lcp_curr_bucket, iter);
     Uint32 loop = 0;
@@ -1270,6 +1300,7 @@ Pgman::process_lcp(Signal* signal)
 	   m_lcp_outstanding < (Uint32) max_count &&
 	   (loop ++ < 32 || iter.bucket == m_lcp_curr_bucket))
     {
+      jam();
       Ptr<Page_entry>& ptr = iter.curr;
       Page_state state = ptr.p->m_state;
       
@@ -1279,6 +1310,7 @@ Pgman::process_lcp(Signal* signal)
           (state & Page_entry::DIRTY) &&
 	  (! (state & Page_entry::LOCKED)))
       {
+        jam();
         if(! (state & Page_entry::BOUND))
         {
           ndbout << ptr << endl;
@@ -1286,21 +1318,24 @@ Pgman::process_lcp(Signal* signal)
         }
         if (state & Page_entry::BUSY)
         {
-	  DBG_LCP(" BUSY" << endl);
+          jam();
+          DBG_LCP(" BUSY" << endl);
           break;  // wait for it
         } 
 	else if (state & Page_entry::PAGEOUT)
         {
-	  DBG_LCP(" PAGEOUT -> state |= LCP" << endl);
+          jam();
+          DBG_LCP(" PAGEOUT -> state |= LCP" << endl);
           set_page_state(ptr, state | Page_entry::LCP);
         }
         else
         {
-	  DBG_LCP(" pageout()" << endl);
+          jam();
+          DBG_LCP(" pageout()" << endl);
           ptr.p->m_state |= Page_entry::LCP;
-	  c_tup->disk_page_unmap_callback(0,
-					  ptr.p->m_real_page_i, 
-					  ptr.p->m_dirty_count);
+          c_tup->disk_page_unmap_callback(0,
+                                          ptr.p->m_real_page_i, 
+                                          ptr.p->m_dirty_count);
           pageout(signal, ptr);
         }
         ptr.p->m_last_lcp = m_last_lcp;
@@ -1308,7 +1343,8 @@ Pgman::process_lcp(Signal* signal)
       }
       else
       {
-	DBG_LCP(" NOT DIRTY" << endl);
+        jam();
+        DBG_LCP(" NOT DIRTY" << endl);
       }	
       pl_hash.next(iter);
     }
@@ -1322,7 +1358,9 @@ Pgman::process_lcp(Signal* signal)
     Page_sublist& pl = *m_page_sublist[Page_entry::SL_LOCKED];
     if (pl.first(ptr))
     {
+      jam();
       process_lcp_locked(signal, ptr);
+      return LS_LCP_LOCKED;
     }
     else
     {
@@ -1335,11 +1373,17 @@ Pgman::process_lcp(Signal* signal)
       }
       signal->theData[0] = m_end_lcp_req.senderData;
       sendSignal(m_end_lcp_req.senderRef, GSN_END_LCP_CONF, signal, 1, JBB);
+      return LS_LCP_OFF;
     }
-    return false;
+  }
+
+  if (m_lcp_outstanding >= (Uint32) max_count)
+  {
+    jam();
+    return LS_LCP_MAX_LCP_OUTSTANDING;
   }
   
-  return true;
+  return LS_LCP_ON;
 }
 
 void
@@ -1546,6 +1590,12 @@ Pgman::fswriteconf(Signal* signal, Ptr<Page_entry> ptr)
   
   set_page_state(ptr, state);
   do_busy_loop(signal, true);
+
+  if (m_lcp_state == LS_LCP_MAX_LCP_OUTSTANDING)
+  {
+    jam();
+    do_lcp_loop(signal);
+  }
 }
 
 // file system interface
@@ -1654,7 +1704,7 @@ Pgman::execFSWRITEREF(Signal* signal)
 // client methods
 
 int
-Pgman::get_page(Signal* signal, Ptr<Page_entry> ptr, Page_request page_req)
+Pgman::get_page_no_lirs(Signal* signal, Ptr<Page_entry> ptr, Page_request page_req)
 {
   jamEntry();
 #ifdef VM_TRACE
@@ -1695,16 +1745,6 @@ Pgman::get_page(Signal* signal, Ptr<Page_entry> ptr, Page_request page_req)
     jam();
   }
 
-  // update LIRS
-  if (! (state & Page_entry::LOCKED) &&
-      ! (req_flags & Page_request::CORR_REQ))
-  {
-    jam();
-    set_page_state(ptr, state);
-    lirs_reference(ptr);
-    state = ptr.p->m_state;
-  }
-
   const Page_state LOCKED = Page_entry::LOCKED | Page_entry::MAPPED;
   if ((state & LOCKED) == LOCKED && 
       ! (req_flags & Page_request::UNLOCK_PAGE))
@@ -1791,8 +1831,6 @@ Pgman::get_page(Signal* signal, Ptr<Page_entry> ptr, Page_request page_req)
   ptr.p->m_busy_count += busy_count;
   ptr.p->m_dirty_count += !!(req_flags & DIRTY_FLAGS);
   set_page_state(ptr, state);
-  
-  do_busy_loop(signal, true);
 
 #ifdef VM_TRACE
   debugOut << "PGMAN: " << req_ptr << endl;
@@ -1801,6 +1839,37 @@ Pgman::get_page(Signal* signal, Ptr<Page_entry> ptr, Page_request page_req)
   return 0;
 }
 
+int
+Pgman::get_page(Signal* signal, Ptr<Page_entry> ptr, Page_request page_req)
+{
+  int i = get_page_no_lirs(signal, ptr, page_req);
+  if (unlikely(i == -1))
+  {
+    jam();
+    return -1;
+  }
+
+  Uint32 req_flags = page_req.m_flags;
+  Page_state state = ptr.p->m_state;
+
+  // update LIRS
+  if (! (state & Page_entry::LOCKED) &&
+      ! (req_flags & Page_request::CORR_REQ))
+  {
+    jam();
+    lirs_reference(ptr);
+  }
+
+  // start processing if request was queued
+  if (i == 0)
+  {
+    jam();
+    do_busy_loop(signal, true);
+  }
+
+  return i;
+}
+
 void
 Pgman::update_lsn(Ptr<Page_entry> ptr, Uint32 block, Uint64 lsn)
 {
@@ -1926,9 +1995,6 @@ Pgman::drop_page(Ptr<Page_entry> ptr)
   Page_state state = ptr.p->m_state;
   if (! (state & (Page_entry::PAGEIN | Page_entry::PAGEOUT)))
   {
-    ndbrequire(state & Page_entry::BOUND);
-    ndbrequire(state & Page_entry::MAPPED);
-
     if (state & Page_entry::ONSTACK)
     {
       jam();
@@ -1938,9 +2004,13 @@ Pgman::drop_page(Ptr<Page_entry> ptr)
       if (at_bottom)
       {
         jam();
-        ndbassert(state & Page_entry::HOT);
         lirs_stack_prune();
       }
+      if (state & Page_entry::HOT)
+      {
+        jam();
+        state &= ~ Page_entry::HOT;
+      }
     }
 
     if (state & Page_entry::ONQUEUE)
@@ -1950,12 +2020,37 @@ Pgman::drop_page(Ptr<Page_entry> ptr)
       state &= ~ Page_entry::ONQUEUE;
     }
 
-    ndbassert(ptr.p->m_real_page_i != RNIL);
-    if (ptr.p->m_real_page_i != RNIL)
+    if (state & Page_entry::BUSY)
     {
       jam();
+      state &= ~ Page_entry::BUSY;
+    }
+
+    if (state & Page_entry::DIRTY)
+    {
+      jam();
+      state &= ~ Page_entry::DIRTY;
+    }
+
+    if (state & Page_entry::EMPTY)
+    {
+      jam();
+      state &= ~ Page_entry::EMPTY;
+    }
+
+    if (state & Page_entry::MAPPED)
+    {
+      jam();
+      state &= ~ Page_entry::MAPPED;
+    }
+
+    if (state & Page_entry::BOUND)
+    {
+      jam();
+      ndbrequire(ptr.p->m_real_page_i != RNIL);
       release_cache_page(ptr.p->m_real_page_i);
       ptr.p->m_real_page_i = RNIL;
+      state &= ~ Page_entry::BOUND;
     }
 
     set_page_state(ptr, state);
@@ -1974,6 +2069,8 @@ Pgman::drop_page(Ptr<Page_entry> ptr)
 void
 Pgman::verify_page_entry(Ptr<Page_entry> ptr)
 {
+  Page_stack& pl_stack = m_page_stack;
+
   Uint32 ptrI = ptr.i;
   Page_state state = ptr.p->m_state;
 
@@ -1996,6 +2093,10 @@ Pgman::verify_page_entry(Ptr<Page_entry> ptr)
   // hot entry must be on stack
   ndbrequire(! is_hot || on_stack || dump_page_lists(ptrI));
 
+  // stack bottom is hot
+  bool at_bottom = on_stack && ! pl_stack.hasPrev(ptr);
+  ndbrequire(! at_bottom || is_hot || dump_page_lists(ptrI));
+
   bool on_queue = state & Page_entry::ONQUEUE;
   // hot entry is not on queue
   ndbrequire(! is_hot || ! on_queue || dump_page_lists(ptrI));
@@ -2007,9 +2108,12 @@ Pgman::verify_page_entry(Ptr<Page_entry> ptr)
   // entries waiting to enter queue
   bool to_queue = ! is_locked && ! is_hot && ! is_bound && has_req;
 
-  // page is either LOCKED or under LIRS
+  // page is about to be released
+  bool to_release = (state == 0);
+
+  // page is either LOCKED or under LIRS or about to be released
   bool is_lirs = on_stack || to_queue || on_queue;
-  ndbrequire(is_locked == ! is_lirs || dump_page_lists(ptrI));
+  ndbrequire(to_release || is_locked == ! is_lirs || dump_page_lists(ptrI));
 
   bool pagein = state & Page_entry::PAGEIN;
   bool pageout = state & Page_entry::PAGEOUT;
@@ -2019,19 +2123,19 @@ Pgman::verify_page_entry(Ptr<Page_entry> ptr)
   Uint32 no = get_sublist_no(state);
   switch (no) {
   case Page_entry::SL_BIND:
-    ndbrequire(! pagein && ! pageout || dump_page_lists(ptrI));
+    ndbrequire((! pagein && ! pageout) || dump_page_lists(ptrI));
     break;
   case Page_entry::SL_MAP:
-    ndbrequire(! pagein && ! pageout || dump_page_lists(ptrI));
+    ndbrequire((! pagein && ! pageout) || dump_page_lists(ptrI));
     break;
   case Page_entry::SL_MAP_IO:
-    ndbrequire(pagein && ! pageout || dump_page_lists(ptrI));
+    ndbrequire((pagein && ! pageout) || dump_page_lists(ptrI));
     break;
   case Page_entry::SL_CALLBACK:
-    ndbrequire(! pagein && ! pageout || dump_page_lists(ptrI));
+    ndbrequire((! pagein && ! pageout) || dump_page_lists(ptrI));
     break;
   case Page_entry::SL_CALLBACK_IO:
-    ndbrequire(! pagein && pageout || dump_page_lists(ptrI));
+    ndbrequire((! pagein && pageout) || dump_page_lists(ptrI));
     break;
   case Page_entry::SL_BUSY:
     break;
@@ -2041,6 +2145,9 @@ Pgman::verify_page_entry(Ptr<Page_entry> ptr)
     break;
   case Page_entry::SL_OTHER:
     break;
+  case ZNIL:
+    ndbrequire(to_release || dump_page_lists(ptrI));
+    break;
   default:
     ndbrequire(false || dump_page_lists(ptrI));
     break;
@@ -2050,122 +2157,127 @@ Pgman::verify_page_entry(Ptr<Page_entry> ptr)
 void
 Pgman::verify_page_lists()
 {
+  const Stats& stats = m_stats;
+  const Param& param = m_param;
   Page_hashlist& pl_hash = m_page_hashlist;
   Page_stack& pl_stack = m_page_stack;
   Page_queue& pl_queue = m_page_queue;
   Ptr<Page_entry> ptr;
 
-  Uint32 stack_count = 0;
-  Uint32 queue_count = 0;
-  Uint32 queuewait_count = 0;
-  Uint32 locked_bound_count = 0;
+  Uint32 is_locked = 0;
+  Uint32 is_bound = 0;
+  Uint32 is_mapped = 0;
+  Uint32 is_hot = 0;
+  Uint32 on_stack = 0;
+  Uint32 on_queue = 0;
+  Uint32 to_queue = 0;
 
   Page_hashlist::Iterator iter;
   pl_hash.next(0, iter);
   while (iter.curr.i != RNIL)
   {
-    verify_page_entry(iter.curr);
+    ptr = iter.curr;
+    Page_state state = ptr.p->m_state;
+    // (state == 0) occurs only within a time-slice
+    ndbrequire(state != 0);
+    verify_page_entry(ptr);
 
-    Page_state state = iter.curr.p->m_state;
+    if (state & Page_entry::LOCKED)
+      is_locked++;
+    if (state & Page_entry::BOUND)
+      is_bound++;
+    if (state & Page_entry::MAPPED)
+      is_mapped++;
+    if (state & Page_entry::HOT)
+      is_hot++;
     if (state & Page_entry::ONSTACK)
-      stack_count++;
+      on_stack++;
     if (state & Page_entry::ONQUEUE)
-      queue_count++;
+      on_queue++;
     if (! (state & Page_entry::LOCKED) &&
         ! (state & Page_entry::HOT) &&
         (state & Page_entry::REQUEST) &&
         ! (state & Page_entry::BOUND))
-      queuewait_count++;
-    if (state & Page_entry::LOCKED &&
-        state & Page_entry::BOUND)
-      locked_bound_count++;
+      to_queue++;
     pl_hash.next(iter);
   }
 
-  ndbrequire(stack_count == pl_stack.count() || dump_page_lists());
-  ndbrequire(queue_count == pl_queue.count() || dump_page_lists());
-
-  Uint32 hot_count = 0;
-  Uint32 hot_bound_count = 0;
-  Uint32 cold_bound_count = 0;
-  Uint32 stack_request_count = 0;
-  Uint32 queue_request_count = 0;
-
-  Uint32 i1 = RNIL;
   for (pl_stack.first(ptr); ptr.i != RNIL; pl_stack.next(ptr))
   {
-    ndbrequire(i1 != ptr.i);
-    i1 = ptr.i;
     Page_state state = ptr.p->m_state;
-    ndbrequire(state & Page_entry::ONSTACK || dump_page_lists());
+    ndbrequire(state & Page_entry::ONSTACK || dump_page_lists(ptr.i));
     if (! pl_stack.hasPrev(ptr))
-      ndbrequire(state & Page_entry::HOT || dump_page_lists());
-    if (state & Page_entry::HOT) {
-      hot_count++;
-      if (state & Page_entry::BOUND)
-        hot_bound_count++;
+    {
+      ndbrequire(state & Page_entry::HOT || dump_page_lists(ptr.i));
     }
-    if (state & Page_entry::REQUEST)
-      stack_request_count++;
   }
 
-  Uint32 i2 = RNIL;
   for (pl_queue.first(ptr); ptr.i != RNIL; pl_queue.next(ptr))
   {
-    ndbrequire(i2 != ptr.i);
-    i2 = ptr.i;
     Page_state state = ptr.p->m_state;
-    ndbrequire(state & Page_entry::ONQUEUE || dump_page_lists());
-    ndbrequire(state & Page_entry::BOUND || dump_page_lists());
-    cold_bound_count++;
-    if (state & Page_entry::REQUEST)
-      queue_request_count++;
+    ndbrequire(state & Page_entry::ONQUEUE || dump_page_lists(ptr.i));
+    ndbrequire(state & Page_entry::BOUND || dump_page_lists(ptr.i));
+    ndbrequire(! (state & Page_entry::HOT) || dump_page_lists(ptr.i));
   }
 
-  Uint32 tot_bound_count =
-    locked_bound_count + hot_bound_count + cold_bound_count;
-  ndbrequire(m_stats.m_num_pages == tot_bound_count || dump_page_lists());
+  ndbrequire(is_bound == stats.m_num_pages || dump_page_lists());
+  ndbrequire(is_hot == stats.m_num_hot_pages || dump_page_lists());
+  ndbrequire(on_stack == pl_stack.count() || dump_page_lists());
+  ndbrequire(on_queue == pl_queue.count() || dump_page_lists());
 
   Uint32 k;
   Uint32 entry_count = 0;
-
+  char sublist_info[200] = "";
   for (k = 0; k < Page_entry::SUBLIST_COUNT; k++)
   {
     const Page_sublist& pl = *m_page_sublist[k];
     for (pl.first(ptr); ptr.i != RNIL; pl.next(ptr))
-    {
-      ndbrequire(get_sublist_no(ptr.p->m_state) == k || dump_page_lists());
-      entry_count++;
-    }
+      ndbrequire(get_sublist_no(ptr.p->m_state) == k || dump_page_lists(ptr.i));
+    entry_count += pl.count();
+    sprintf(sublist_info + strlen(sublist_info),
+            " %s:%u", get_sublist_name(k), pl.count());
   }
-
   ndbrequire(entry_count == pl_hash.count() || dump_page_lists());
 
-  debugOut << "PGMAN: loop"
-           << " stats=" << m_stats_loop_on
-           << " busy=" << m_busy_loop_on
-           << " cleanup=" << m_cleanup_loop_on
-           << " lcp=" << m_lcp_loop_on << endl;
+  Uint32 hit_pct = 0;
+  char hit_pct_str[20];
+  if (stats.m_page_hits + stats.m_page_faults != 0)
+    hit_pct = 10000 * stats.m_page_hits /
+              (stats.m_page_hits + stats.m_page_faults);
+  sprintf(hit_pct_str, "%u.%02u", hit_pct / 100, hit_pct % 100);
 
-  debugOut << "PGMAN:"
-           << " entry:" << pl_hash.count()
-           << " cache:" << m_stats.m_num_pages
-           << "(" << locked_bound_count << "L)"
-           << " stack:" << pl_stack.count()
-           << " hot:" << hot_count
-           << " hot_bound:" << hot_bound_count
-           << " stack_request:" << stack_request_count
-           << " queue:" << pl_queue.count()
-           << " queue_request:" << queue_request_count
-           << " queuewait:" << queuewait_count << endl;
+  if (! debugFlag && debugSummaryFlag)
+    open_debug_file(1);
 
-  debugOut << "PGMAN:";
-  for (k = 0; k < Page_entry::SUBLIST_COUNT; k++)
-  {
-    const Page_sublist& pl = *m_page_sublist[k];
-    debugOut << " " << get_sublist_name(k) << ":" << pl.count();
-  }
-  debugOut << endl;
+  debugOut
+    << "PGMAN: loop"
+    << " stats:" << m_stats_loop_on
+    << " busy:" << m_busy_loop_on
+    << " cleanup:" << m_cleanup_loop_on
+    << " lcp:" << Uint32(m_lcp_state) << endl;
+
+  debugOut
+    << "PGMAN:"
+    << " entries:" << pl_hash.count()
+    << " pages:" << stats.m_num_pages << "/" << param.m_max_pages
+    << " mapped:" << is_mapped
+    << " hot:" << is_hot
+    << " io:" << stats.m_current_io_waits << "/" << param.m_max_io_waits
+    << " hit pct:" << hit_pct_str << endl;
+
+  debugOut
+    << "PGMAN:"
+    << " locked:" << is_locked
+    << " stack:" << pl_stack.count()
+    << " queue:" << pl_queue.count()
+    << " to queue:" << to_queue << endl;
+
+  debugOut
+    << "PGMAN:"
+    << sublist_info << endl;
+
+  if (! debugFlag && debugSummaryFlag)
+    open_debug_file(0);
 }
 
 void
@@ -2188,56 +2300,36 @@ Pgman::dump_page_lists(Uint32 ptrI)
   if (! debugFlag)
     open_debug_file(1);
 
-  debugOut << "PGMAN: page list dump" << endl;
+  debugOut << "PGMAN: page list dump" << "\n";
   if (ptrI != RNIL)
-    debugOut << "PGMAN: error on PE [" << ptrI << "]" << endl;
+    debugOut << "PGMAN: error on PE [" << ptrI << "]" << "\n";
 
-  Page_hashlist& pl_hash = m_page_hashlist;
   Page_stack& pl_stack = m_page_stack;
   Page_queue& pl_queue = m_page_queue;
   Ptr<Page_entry> ptr;
   Uint32 n;
-  char buf[40];
 
-  debugOut << "hash:" << endl;
-  Page_hashlist::Iterator iter;
-  pl_hash.next(0, iter);
-  n = 0;
-  while (iter.curr.i != RNIL)
-  {
-    sprintf(buf, "%03d", n++);
-    debugOut << buf << " " << iter.curr << endl;
-    pl_hash.next(iter);
-  }
-
-  debugOut << "stack:" << endl;
+  debugOut << "stack:" << "\n";
   n = 0;
   for (pl_stack.first(ptr); ptr.i != RNIL; pl_stack.next(ptr))
-  {
-    sprintf(buf, "%03d", n++);
-    debugOut << buf << " " << ptr << endl;
-  }
+    debugOut << n++ << " " << ptr << "\n";
 
-  debugOut << "queue:" << endl;
+  debugOut << "queue:" << "\n";
   n = 0;
   for (pl_queue.first(ptr); ptr.i != RNIL; pl_queue.next(ptr))
-  {
-    sprintf(buf, "%03d", n++);
-    debugOut << buf << " " << ptr << endl;
-  }
+    debugOut << n++ << " " << ptr << "\n";
 
   Uint32 k;
   for (k = 0; k < Page_entry::SUBLIST_COUNT; k++)
   {
-    debugOut << get_sublist_name(k) << ":" << endl;
+    debugOut << get_sublist_name(k) << ":" << "\n";
     const Page_sublist& pl = *m_page_sublist[k];
+    n = 0;
     for (pl.first(ptr); ptr.i != RNIL; pl.next(ptr))
-    {
-      sprintf(buf, "%03d", n++);
-    debugOut << buf << " " << ptr << endl;
-    }
+      debugOut << n++ << " " << ptr << "\n";
   }
 
+  debugOut.flushline();
   if (! debugFlag)
     open_debug_file(0);
 
@@ -2257,9 +2349,9 @@ Pgman::get_sublist_name(Uint32 list_no)
   case Page_entry::SL_MAP_IO:
     return "map_io";
   case Page_entry::SL_CALLBACK:
-    return "callback";
+    return "cb";
   case Page_entry::SL_CALLBACK_IO:
-    return "callback_io";
+    return "cb_io";
   case Page_entry::SL_BUSY:
     return "busy";
   case Page_entry::SL_LOCKED:
@@ -2404,8 +2496,9 @@ Pgman::execDUMP_STATE_ORD(Signal* signal)
   if (signal->theData[0] == 11000 && signal->getLength() == 2)
   {
     Uint32 flag = signal->theData[1];
-    open_debug_file(flag);
-    debugFlag = flag;
+    debugFlag = flag & 1;
+    debugSummaryFlag = flag & 2;
+    open_debug_file(debugFlag);
   }
 #endif
 
@@ -2502,6 +2595,11 @@ Pgman::execDUMP_STATE_ORD(Signal* signal)
   {
     SET_ERROR_INSERT_VALUE(11008);
   }
+
+  if (signal->theData[0] == 11009)
+  {
+    SET_ERROR_INSERT_VALUE(11009);
+  }
 }
 
 // page cache client
