@@ -1,4 +1,6 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2005-2008 MySQL AB, 2010 Sun Microsystems, Inc.
+   Use is subject to license terms.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +13,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include "restore.hpp"
 #include <signaldata/FsRef.hpp>
@@ -359,7 +362,7 @@ Restore::open_file(Signal* signal, FilePtr file_ptr)
 {
   FsOpenReq * req = (FsOpenReq *)signal->getDataPtrSend();
   req->userReference = reference();
-  req->fileFlags = FsOpenReq::OM_READONLY ;
+  req->fileFlags = FsOpenReq::OM_READONLY | FsOpenReq::OM_GZ;
   req->userPointer = file_ptr.i;
   
   FsOpenReq::setVersion(req->fileNumber, 5);
@@ -484,7 +487,7 @@ Restore::restore_next(Signal* signal, FilePtr file_ptr)
       
       if (unlikely((status & File:: FILE_THREAD_RUNNING) == 0))
       {
-	ndbrequire(false);
+        crash_during_restore(file_ptr, __LINE__, 0);
       }
       len= 0;
       break;
@@ -712,12 +715,9 @@ Restore::execFSCLOSECONF(Signal * signal)
 
   if(file_ptr.p->m_outstanding_operations == 0)
   {
-    RestoreLcpConf* rep= (RestoreLcpConf*)signal->getDataPtrSend();
-    rep->senderData= file_ptr.p->m_sender_data;
-    sendSignal(file_ptr.p->m_sender_ref, 
-	       GSN_RESTORE_LCP_CONF, signal, 
-	       RestoreLcpConf::SignalLength, JBB);
-    release_file(file_ptr);
+    jam();
+    restore_lcp_conf(signal, file_ptr);
+    return;
   }
 }
 
@@ -734,7 +734,8 @@ Restore::parse_file_header(Signal* signal,
     return;
   }
   
-  if (check_file_version(signal, ntohl(fh->NdbVersion)))
+  file_ptr.p->m_lcp_version = ntohl(fh->BackupVersion);
+  if (check_file_version(signal, ntohl(fh->BackupVersion)))
   {
     parse_error(signal, file_ptr, __LINE__, ntohl(fh->NdbVersion));
     return;
@@ -941,100 +942,117 @@ Restore::parse_record(Signal* signal, FilePtr file_ptr,
   data += 1;
   const Uint32* const dataStart = data;
 
-  Uint32 *keyData = key_start;
-  Uint32 *attrData = attr_start;
-  union {
-    Column c;
-    Uint32 _align[sizeof(Column)/sizeof(Uint32)];
-  };
   bool disk = false;
   bool rowid = false;
   bool gci = false;
-  Uint32 tableId = file_ptr.p->m_table_id;
-
-  Uint64 gci_val;
+  Uint32 keyLen;
+  Uint32 attrLen;
   Local_key rowid_val;
-  columns.first(it);
-  while(!it.isNull())
-  {
-    _align[0] = *it.data; ndbrequire(columns.next(it));
-    _align[1] = *it.data; columns.next(it);
-    
-    if (c.m_id == AttributeHeader::ROWID)
-    {
-      rowid_val.m_page_no = data[0];
-      rowid_val.m_page_idx = data[1];
-      data += 2;
-      rowid = true;
-      continue;
-    }
-
-    if (c.m_id == AttributeHeader::ROW_GCI)
-    {
-      memcpy(&gci_val, data, 8);
-      data += 2;
-      gci = true;
-      continue;
-    }
-    
-    if (! (c.m_flags & (Column::COL_VAR | Column::COL_NULL)))
-    {
-      ndbrequire(data < dataStart + len);
-
-      if(c.m_flags & Column::COL_KEY)
-      {
-        memcpy(keyData, data, 4*c.m_size);
-        keyData += c.m_size;
-      } 
-
-      AttributeHeader::init(attrData++, c.m_id, c.m_size << 2);
-      memcpy(attrData, data, 4*c.m_size);
-      attrData += c.m_size;
-      data += c.m_size;
-    }
-
-    if(c.m_flags & Column::COL_DISK)
-      disk= true;
-  }
-
-  // second part is data driven
-  while (data + 2 < dataStart + len) {
-    Uint32 sz= ntohl(*data); data++;
-    Uint32 id= ntohl(*data); data++; // column_no
-
-    ndbrequire(columns.position(it, 2 * id));
-
-    _align[0] = *it.data; ndbrequire(columns.next(it));
-    _align[1] = *it.data;
-
-    Uint32 sz32 = (sz + 3) >> 2;
-    ndbassert(c.m_flags & (Column::COL_VAR | Column::COL_NULL));
-    if (c.m_flags & Column::COL_KEY)
-    {
-      memcpy(keyData, data, 4 * sz32);
-      keyData += sz32;
-    }
-
-    AttributeHeader::init(attrData++, c.m_id, sz);
-    memcpy(attrData, data, sz);
-
-    attrData += sz32;
-    data += sz32;
-  }
-
-  ndbrequire(data == dataStart + len - 1);
-
-  ndbrequire(disk == false); // Not supported...
-  ndbrequire(rowid == true);
-  Uint32 keyLen = keyData - key_start;
-  Uint32 attrLen = attrData - attr_start;
-  LqhKeyReq * req = (LqhKeyReq *)signal->getDataPtrSend();
-  
+  Uint64 gci_val;
+  Uint32 tableId = file_ptr.p->m_table_id;
   const KeyDescriptor* desc = g_key_descriptor_pool.getPtr(tableId);
-  if (desc->noOfKeyAttr != desc->noOfVarKeys)
+
+  if (likely(file_ptr.p->m_lcp_version >= NDBD_RAW_LCP))
   {
-    reorder_key(desc, key_start, keyLen);
+    rowid = true;
+    rowid_val.m_page_no = data[0];
+    rowid_val.m_page_idx = data[1];
+    keyLen = c_tup->read_lcp_keys(tableId, data+2, len - 3, key_start);
+
+    AttributeHeader::init(attr_start, AttributeHeader::READ_LCP, 4*(len - 3));
+    memcpy(attr_start + 1, data + 2, 4 * (len - 3));
+    attrLen = 1 + len - 3;
   }
+  else
+  {
+    Uint32 *keyData = key_start;
+    Uint32 *attrData = attr_start;
+    union {
+      Column c;
+      Uint32 _align[sizeof(Column)/sizeof(Uint32)];
+    };
+    
+    columns.first(it);
+    while(!it.isNull())
+    {
+      _align[0] = *it.data; ndbrequire(columns.next(it));
+      _align[1] = *it.data; columns.next(it);
+
+      if (c.m_id == AttributeHeader::ROWID)
+      {
+        rowid_val.m_page_no = data[0];
+        rowid_val.m_page_idx = data[1];
+        data += 2;
+        rowid = true;
+        continue;
+      }
+
+      if (c.m_id == AttributeHeader::ROW_GCI)
+      {
+        memcpy(&gci_val, data, 8);
+        data += 2;
+        gci = true;
+        continue;
+      }
+
+      if (! (c.m_flags & (Column::COL_VAR | Column::COL_NULL)))
+      {
+        ndbrequire(data < dataStart + len);
+
+        if(c.m_flags & Column::COL_KEY)
+        {
+          memcpy(keyData, data, 4*c.m_size);
+          keyData += c.m_size;
+        }
+
+        AttributeHeader::init(attrData++, c.m_id, c.m_size << 2);
+        memcpy(attrData, data, 4*c.m_size);
+        attrData += c.m_size;
+        data += c.m_size;
+      }
+
+      if(c.m_flags & Column::COL_DISK)
+        disk= true;
+    }
+
+    // second part is data driven
+    while (data + 2 < dataStart + len) {
+      Uint32 sz= ntohl(*data); data++;
+      Uint32 id= ntohl(*data); data++; // column_no
+
+      ndbrequire(columns.position(it, 2 * id));
+
+      _align[0] = *it.data; ndbrequire(columns.next(it));
+      _align[1] = *it.data;
+
+      Uint32 sz32 = (sz + 3) >> 2;
+      ndbassert(c.m_flags & (Column::COL_VAR | Column::COL_NULL));
+      if (c.m_flags & Column::COL_KEY)
+      {
+        memcpy(keyData, data, 4 * sz32);
+        keyData += sz32;
+      }
+
+      AttributeHeader::init(attrData++, c.m_id, sz);
+      memcpy(attrData, data, sz);
+
+      attrData += sz32;
+      data += sz32;
+    }
+
+    ndbrequire(data == dataStart + len - 1);
+
+    ndbrequire(disk == false); // Not supported...
+    ndbrequire(rowid == true);
+    keyLen = keyData - key_start;
+    attrLen = attrData - attr_start;
+    if (desc->noOfKeyAttr != desc->noOfVarKeys)
+    {
+      reorder_key(desc, key_start, keyLen);
+    }
+  }
+  
+  LqhKeyReq * req = (LqhKeyReq *)signal->getDataPtrSend();
   
   Uint32 hashValue;
   if (g_key_descriptor_pool.getPtr(tableId)->hasCharAttr)
@@ -1121,12 +1139,12 @@ Restore::reorder_key(const KeyDescriptor* desc,
       src += sz;
       break;
     case NDB_ARRAYTYPE_SHORT_VAR:
-      sz = (1 + ((char*)var)[0] + 3) >> 2;
+      sz = (1 + ((Uint8*)var)[0] + 3) >> 2;
       memcpy(dst, var, 4 * sz);
       var += sz;
       break;
     case NDB_ARRAYTYPE_MEDIUM_VAR:
-      sz = (2 + ((char*)var)[0] +  256*((char*)var)[1] + 3) >> 2;
+      sz = (2 + ((Uint8*)var)[0] +  256*((Uint8*)var)[1] + 3) >> 2;
       memcpy(dst, var, 4 * sz);
       var += sz;
       break;
@@ -1160,18 +1178,32 @@ Restore::execLQHKEYREF(Signal* signal)
   LqhKeyRef* ref = (LqhKeyRef*)signal->getDataPtr();
   m_file_pool.getPtr(file_ptr, ref->connectPtr);
   
+  crash_during_restore(file_ptr, __LINE__, ref->errorCode);
+  ndbrequire(false);
+}
+
+void
+Restore::crash_during_restore(FilePtr file_ptr, Uint32 line, Uint32 errCode)
+{
   char buf[255], name[100];
   BaseString::snprintf(name, sizeof(name), "%u/T%dF%d",
 		       file_ptr.p->m_lcp_no,
 		       file_ptr.p->m_table_id,
 		       file_ptr.p->m_fragment_id);
   
-  BaseString::snprintf(buf, sizeof(buf),
-		       "Error %d during restore of  %s",
-		       ref->errorCode, name);
-  
+  if (errCode)
+  {
+    BaseString::snprintf(buf, sizeof(buf),
+                         "Error %d (line: %u) during restore of  %s",
+                         errCode, line, name);
+  }
+  else
+  {
+    BaseString::snprintf(buf, sizeof(buf),
+                         "Error (line %u) during restore of  %s",
+                         line, name);
+  }
   progError(__LINE__, NDBD_EXIT_INVALID_LCP_FILE, buf);  
-  ndbrequire(false);
 }
 
 void
@@ -1186,15 +1218,39 @@ Restore::execLQHKEYCONF(Signal* signal)
   file_ptr.p->m_rows_restored++;
   if(file_ptr.p->m_outstanding_operations == 0 && file_ptr.p->m_fd == RNIL)
   {
-    RestoreLcpConf* rep= (RestoreLcpConf*)signal->getDataPtrSend();
-    rep->senderData= file_ptr.p->m_sender_data;
-    sendSignal(file_ptr.p->m_sender_ref, 
-	       GSN_RESTORE_LCP_CONF, signal, 
-	       RestoreLcpConf::SignalLength, JBB);
-    release_file(file_ptr);
+    jam();
+    restore_lcp_conf(signal, file_ptr);
+    return;
   }
 }
 
+void
+Restore::restore_lcp_conf(Signal* signal, FilePtr file_ptr)
+{
+  RestoreLcpConf* rep= (RestoreLcpConf*)signal->getDataPtrSend();
+  rep->senderData= file_ptr.p->m_sender_data;
+  if(file_ptr.p->is_lcp())
+  {
+    /**
+     * Temporary reset DBTUP's #disk attributes on table
+     *
+     * TUP will send RESTORE_LCP_CONF
+     */
+    c_tup->complete_restore_lcp(signal, 
+                                file_ptr.p->m_sender_ref,
+                                file_ptr.p->m_sender_data,
+                                file_ptr.p->m_table_id,
+				file_ptr.p->m_fragment_id);
+  }
+  else
+  {
+    sendSignal(file_ptr.p->m_sender_ref, 
+               GSN_RESTORE_LCP_CONF, signal, 
+               RestoreLcpConf::SignalLength, JBB);
+  }
+  release_file(file_ptr);
+}
+
 void
 Restore::parse_fragment_footer(Signal* signal, FilePtr file_ptr, 
 			       const Uint32 *data, Uint32 len)
@@ -1212,17 +1268,6 @@ Restore::parse_fragment_footer(Signal* signal, FilePtr file_ptr,
     parse_error(signal, file_ptr, __LINE__, ntohl(fh->SectionLength));
     return;
   }
-
-  if(file_ptr.p->is_lcp())
-  {
-    /**
-     * Temporary reset DBTUP's #disk attributes on table
-     */
-    c_tup->complete_restore_lcp(file_ptr.p->m_table_id,
-				file_ptr.p->m_fragment_id);
-  }
-
-  file_ptr.p->m_fragment_id = RNIL;
 }
 
 void
