@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #ifndef SIMULATEDBLOCK_H
 #define SIMULATEDBLOCK_H
@@ -91,11 +93,14 @@ class SimulatedBlock {
   friend class SafeCounterManager;
   friend struct UpgradeStartup;
   friend class AsyncFile;
+  friend class PosixAsyncFile; // FIXME
   friend class Pgman;
   friend class Page_cache_client;
   friend class Lgman;
   friend class Logfile_client;
   friend struct Pool_context;
+  friend struct SectionHandle;
+  friend class LockQueue;
 public:
   friend class BlockComponent;
   virtual ~SimulatedBlock();
@@ -114,6 +119,7 @@ protected:
   void addRecSignalImpl(GlobalSignalNumber g, ExecFunction fun, bool f =false);
   void installSimulatedBlockFunctions();
   ExecFunction theExecArray[MAX_GSN+1];
+
   void initCommon();
 public:
   /**
@@ -158,6 +164,20 @@ protected:
                   Signal* signal, 
 		  Uint32 length, 
 		  JobBufferLevel jbuf,
+		  SectionHandle* sections) const;
+
+  void sendSignal(NodeReceiverGroup rg,
+		  GlobalSignalNumber gsn,
+                  Signal* signal,
+		  Uint32 length,
+		  JobBufferLevel jbuf,
+		  SectionHandle* sections) const;
+
+  void sendSignal(BlockReference ref,
+		  GlobalSignalNumber gsn,
+                  Signal* signal,
+		  Uint32 length,
+		  JobBufferLevel jbuf,
 		  LinearSectionPtr ptr[3],
 		  Uint32 noOfSections) const ;
   
@@ -178,13 +198,54 @@ protected:
                            Uint32 delayInMilliSeconds, 
 			   Uint32 length) const ;
 
+  void sendSignalWithDelay(BlockReference ref,
+			   GlobalSignalNumber gsn,
+                           Signal* signal,
+                           Uint32 delayInMilliSeconds,
+			   Uint32 length,
+			   SectionHandle* sections) const;
+
   void EXECUTE_DIRECT(Uint32 block, 
 		      Uint32 gsn, 
 		      Signal* signal, 
 		      Uint32 len);
   
   class SectionSegmentPool& getSectionSegmentPool();
-  void releaseSections(Signal* signal);
+  void releaseSections(struct SectionHandle&);
+
+  void handle_invalid_sections_in_send_signal(Signal*) const;
+  void handle_lingering_sections_after_execute(Signal*) const;
+  void handle_lingering_sections_after_execute(SectionHandle*) const;
+  void handle_invalid_fragmentInfo(Signal*) const;
+  void handle_send_failed(SendStatus, Signal*) const;
+  void handle_out_of_longsignal_memory(Signal*) const;
+
+  /**
+   * Send routed signals (ONLY LOCALLY)
+   *
+   * NOTE: Only localhost is allowed!
+   */
+  struct RoutePath
+  {
+    Uint32 ref;
+    JobBufferLevel prio;
+  };
+  void sendRoutedSignal(RoutePath path[],
+                        Uint32 pathcnt,      // #hops
+                        Uint32 dst[],        // Final destination(s)
+                        Uint32 dstcnt,       // #final destination(s)
+                        Uint32 gsn,          // Final GSN
+                        Signal*,
+                        Uint32 len,
+                        JobBufferLevel prio, // Final prio
+                        SectionHandle * handle = 0);
+
+
+  /**
+   * Check that signal sent from remote node
+   *   is guaranteed to be correctly serialized wrt to NODE_FAILREP
+   */
+  bool checkNodeFailSequence(Signal*);
 
   /**********************************************************
    * Fragmented signals
@@ -203,6 +264,7 @@ protected:
 			    Signal* signal, 
 			    Uint32 length, 
 			    JobBufferLevel jbuf,
+			    SectionHandle * sections,
 			    Callback & = TheEmptyCallback,
 			    Uint32 messageSize = 240);
 
@@ -211,6 +273,7 @@ protected:
 			    Signal* signal, 
 			    Uint32 length, 
 			    JobBufferLevel jbuf,
+			    SectionHandle * sections,
 			    Callback & = TheEmptyCallback,
 			    Uint32 messageSize = 240);
 
@@ -221,7 +284,7 @@ protected:
 			    JobBufferLevel jbuf,
 			    LinearSectionPtr ptr[3],
 			    Uint32 noOfSections,
-			    Callback &,
+			    Callback & = TheEmptyCallback,
 			    Uint32 messageSize = 240);
 
   void sendFragmentedSignal(NodeReceiverGroup rg, 
@@ -231,9 +294,40 @@ protected:
 			    JobBufferLevel jbuf,
 			    LinearSectionPtr ptr[3],
 			    Uint32 noOfSections,
-			    Callback &,
+			    Callback & = TheEmptyCallback,
 			    Uint32 messageSize = 240);
 
+  /**
+   * simBlockNodeFailure
+   *
+   * Method must be called by blocks that send or receive 
+   * remote Fragmented Signals when they detect a node 
+   * (NDBD or API) failure.
+   * If the block needs to acknowledge or perform further
+   * processing after completing block-level node failure 
+   * handling, it can supply a Callback which will be invoked 
+   * when block-level node failure handling has completed.
+   * Otherwise TheEmptyCallback is used.
+   * If TheEmptyCallback is used, all failure handling is
+   * performed in the current timeslice, to avoid any
+   * races.
+   * 
+   * Parameters
+   *   signal       : Current signal*
+   *   failedNodeId : Node id of failed node
+   *   cb           : Callback to be executed when block-level
+   *                  node failure handling completed.
+   *                  TheEmptyCallback is passed if no further
+   *                  processing is required.
+   * Returns
+   *   Number of 'resources' cleaned up in call.
+   *   Callback return code is total resources cleaned up.
+   *   
+   */
+  Uint32 simBlockNodeFailure(Signal* signal,
+                             Uint32 failedNodeId,
+                             Callback& cb = TheEmptyCallback);
+
   /**********************************************************
    * Fragmented signals structures
    */
@@ -270,11 +364,12 @@ protected:
     
     enum Status {
       SendNotComplete = 0,
-      SendComplete    = 1
+      SendComplete    = 1,
+      SendCancelled   = 2
     };
     Uint8  m_status;
     Uint8  m_prio;
-    Uint16  m_fragInfo;
+    Uint16 m_fragInfo;
     Uint16 m_gsn;
     Uint16 m_messageSize; // Size of each fragment
     Uint32 m_fragmentId;
@@ -303,8 +398,7 @@ protected:
 			 Signal* signal, 
 			 Uint32 length, 
 			 JobBufferLevel jbuf,
-			 LinearSectionPtr ptr[3],
-			 Uint32 noOfSections,
+			 SectionHandle * sections,
 			 Uint32 messageSize = 240);
   
   bool sendFirstFragment(FragmentSendInfo & info,
@@ -313,6 +407,8 @@ protected:
 			 Signal* signal, 
 			 Uint32 length, 
 			 JobBufferLevel jbuf,
+			 LinearSectionPtr ptr[3],
+			 Uint32 noOfSections,
 			 Uint32 messageSize = 240);
   
   /**
@@ -346,12 +442,31 @@ protected:
    * If the cause of the shutdown is known use extradata to add an 
    * errormessage describing the problem
    */
-  void progError(int line, int err_code, const char* extradata=NULL) const ;
+  void progError(int line, int err_code, const char* extradata=NULL) const
+    ATTRIBUTE_NORETURN;
 private:
-  void  signal_error(Uint32, Uint32, Uint32, const char*, int) const ;
+  void  signal_error(Uint32, Uint32, Uint32, const char*, int) const
+    ATTRIBUTE_NORETURN;
   const NodeId         theNodeId;
   const BlockNumber    theNumber;
   const BlockReference theReference;
+
+  Uint32 doNodeFailureCleanup(Signal* signal,
+                              Uint32 failedNodeId,
+                              Uint32 resource,
+                              Uint32 cursor,
+                              Uint32 elementsCleaned,
+                              Callback& cb);
+
+  bool doCleanupFragInfo(Uint32 failedNodeId,
+                         Uint32& cursor,
+                         Uint32& rtUnitsUsed,
+                         Uint32& elementsCleaned);
+
+  bool doCleanupFragSend(Uint32 failedNodeId,
+                         Uint32& cursor,
+                         Uint32& rtUnitsUsed,
+                         Uint32& elementsCleaned);
   
 protected:
   Block_context m_ctx;
@@ -390,11 +505,32 @@ protected:
    */
   void deallocRecord(void **, const char * type, size_t s, size_t n);
   
+  /**
+   * Allocate memory from global pool,
+   *   returns #chunks used
+   *
+   * Typically used by part of code, not converted to use global pool
+   *   directly, but allocates everything during startup
+   */
+  struct AllocChunk
+  {
+    Uint32 ptrI;
+    Uint32 cnt;
+  };
+  Uint32 allocChunks(AllocChunk dst[], Uint32 /* size of dst */ arraysize,
+                     Uint32 /* resource group */ rg,
+                     Uint32 /* no of pages to allocate */ pages,
+                     Uint32 paramId /* for error message if failing */);
+
+  static int sortchunks(const void*, const void*);
+
   /**
    * General info event (sent to cluster log)
    */
-  void infoEvent(const char * msg, ...) const ;
-  void warningEvent(const char * msg, ...) const ;
+  void infoEvent(const char * msg, ...) const
+    ATTRIBUTE_FORMAT(printf, 2, 3);
+  void warningEvent(const char * msg, ...) const
+    ATTRIBUTE_FORMAT(printf, 2, 3);
   
   /**
    * Get node state
@@ -432,6 +568,19 @@ protected:
 			  Uint32 *data, 
 			  const Uint32 keyPaLen[MAX_ATTRIBUTES_IN_INDEX])const;
   
+  /**
+   * if ndbd,
+   *   wakeup main-loop if sleeping on IO
+   * if ndbmtd
+   *   wakeup thread running block-instance
+   */
+  void wakeup();
+
+  /**
+   * setup struct for wakeup
+   */
+  void setup_wakeup();
+
 private:
   NewVARIABLE* NewVarRef;      /* New Base Address Table for block  */
   Uint16       theBATSize;     /* # entries in BAT */
@@ -440,11 +589,6 @@ protected:
   ArrayPool<GlobalPage>& m_global_page_pool;
   ArrayPool<GlobalPage>& m_shared_page_pool;
   
-private:
-  /**
-   * Node state
-   */
-  NodeState theNodeState;
   void execNDB_TAMPER(Signal * signal);
   void execNODE_STATE_REP(Signal* signal);
   void execCHANGE_NODE_STATE_REQ(Signal* signal);
@@ -454,6 +598,13 @@ private:
   void execAPI_START_REP(Signal* signal);
   void execNODE_START_REP(Signal* signal);
 
+  void execLOCAL_ROUTE_ORD(Signal*);
+private:
+  /**
+   * Node state
+   */
+  NodeState theNodeState;
+
   Uint32 c_fragmentIdCounter;
   ArrayPool<FragmentInfo> c_fragmentInfoPool;
   DLHashTable<FragmentInfo> c_fragmentInfoHash;
@@ -462,6 +613,9 @@ private:
   ArrayPool<FragmentSendInfo> c_fragmentSendPool;
   DLList<FragmentSendInfo> c_linearFragmentSendList;
   DLList<FragmentSendInfo> c_segmentedFragmentSendList;
+
+protected:
+  Uint32 debugPrintFragmentCounts();
   
 public: 
   class MutexManager {
@@ -479,9 +633,9 @@ public:
      * core interface
      */
     struct ActiveMutex {
+      ActiveMutex() {}
       Uint32 m_gsn; // state
       Uint32 m_mutexId;
-      Uint32 m_mutexKey;
       Callback m_callback;
       union {
 	Uint32 nextPool;
@@ -498,8 +652,7 @@ public:
     
     void create(Signal*, ActiveMutexPtr&);
     void destroy(Signal*, ActiveMutexPtr&);
-    void lock(Signal*, ActiveMutexPtr&);
-    void trylock(Signal*, ActiveMutexPtr&);
+    void lock(Signal*, ActiveMutexPtr&, Uint32 flags);
     void unlock(Signal*, ActiveMutexPtr&);
     
   private:
@@ -566,9 +719,11 @@ public:
 #endif
 
 #ifdef VM_TRACE
-  Ptr<void> **m_global_variables;
+  Ptr<void> **m_global_variables, **m_global_variables_save;
   void clear_global_variables();
   void init_globals_list(void ** tmp, size_t cnt);
+  void disable_global_variables();
+  void enable_global_variables();
 #endif
 };
 
@@ -581,6 +736,11 @@ SimulatedBlock::executeFunction(GlobalSignalNumber gsn, Signal* signal){
     clear_global_variables();
 #endif
     (this->*f)(signal);
+
+    if (unlikely(signal->header.m_noOfSections))
+    {
+      handle_lingering_sections_after_execute(signal);
+    }
     return;
   }
 
@@ -811,8 +971,7 @@ public:\
 private: \
   void addRecSignal(GlobalSignalNumber gsn, ExecSignalLocal f, bool force = false)
 
-#define BLOCK_CONSTRUCTOR(BLOCK) { initCommon();}
-
+#define BLOCK_CONSTRUCTOR(BLOCK) do { SimulatedBlock::initCommon(); } while(0)
 
 #define BLOCK_FUNCTIONS(BLOCK) \
 void \
@@ -822,5 +981,51 @@ BLOCK::addRecSignal(GlobalSignalNumber gsn, ExecSignalLocal f, bool force){ \
 
 #include "Mutex.hpp"
 
+inline
+SectionHandle::~SectionHandle()
+{
+  if (unlikely(m_cnt))
+  {
+    m_block->handle_lingering_sections_after_execute(this);
+  }
+}
+
+#ifdef ERROR_INSERT
+#define RSS_AP_SNAPSHOT(x) Uint32 rss_##x
+#define RSS_AP_SNAPSHOT_SAVE(x) rss_##x = x.getNoOfFree()
+#define RSS_AP_SNAPSHOT_CHECK(x) ndbrequire(rss_##x == x.getNoOfFree())
+#define RSS_AP_SNAPSHOT_SAVE2(x,y) rss_##x = x.getNoOfFree()+(y)
+#define RSS_AP_SNAPSHOT_CHECK2(x,y) ndbrequire(rss_##x == x.getNoOfFree()+(y))
+
+#define RSS_OP_COUNTER(x) Uint32 x
+#define RSS_OP_COUNTER_INIT(x) x = 0
+#define RSS_OP_ALLOC(x) x ++
+#define RSS_OP_FREE(x) x --
+#define RSS_OP_ALLOC_X(x,n) x += n
+#define RSS_OP_FREE_X(x,n) x -= n
+
+#define RSS_OP_SNAPSHOT(x) Uint32 rss_##x
+#define RSS_OP_SNAPSHOT_SAVE(x) rss_##x = x
+#define RSS_OP_SNAPSHOT_CHECK(x) ndbrequire(rss_##x == x)
+#else
+#define RSS_AP_SNAPSHOT(x) struct rss_dummy0_##x { int dummy; }
+#define RSS_AP_SNAPSHOT_SAVE(x)
+#define RSS_AP_SNAPSHOT_CHECK(x)
+#define RSS_AP_SNAPSHOT_SAVE2(x,y)
+#define RSS_AP_SNAPSHOT_CHECK2(x,y)
+
+#define RSS_OP_COUNTER(x) struct rss_dummy1_##x { int dummy; }
+#define RSS_OP_COUNTER_INIT(x)
+#define RSS_OP_ALLOC(x)
+#define RSS_OP_FREE(x)
+#define RSS_OP_ALLOC_X(x,n)
+#define RSS_OP_FREE_X(x,n)
+
+#define RSS_OP_SNAPSHOT(x) struct rss_dummy2_##x { int dummy; }
+#define RSS_OP_SNAPSHOT_SAVE(x)
+#define RSS_OP_SNAPSHOT_CHECK(x)
+
+#endif
+
 #endif
 
