@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 
 
@@ -145,15 +147,17 @@ Ndb::NDB_connect(Uint32 tNode)
 //***************************************************************************
   
   int	         tReturnCode;
-  TransporterFacade *tp = theImpl->m_transporter_facade;
 
   DBUG_ENTER("Ndb::NDB_connect");
 
-  bool nodeAvail = tp->get_node_alive(tNode);
-  if(nodeAvail == false){
-    DBUG_RETURN(0);
+  {
+    TransporterFacade *tp = theImpl->m_transporter_facade;
+    if (tp->get_node_stopping(tNode))
+    {
+      DBUG_RETURN(0);
+    }
   }
-  
+
   NdbTransaction * tConArray = theConnectionArray[tNode];
   if (tConArray != NULL) {
     DBUG_RETURN(2);
@@ -195,19 +199,31 @@ Ndb::NDB_connect(Uint32 tNode)
     tNdbCon->theNext = tPrevFirst;
     DBUG_RETURN(1);
   } else {
-    releaseNdbCon(tNdbCon);
 //****************************************************************************
 // Unsuccessful connect is indicated by 3.
 //****************************************************************************
     DBUG_PRINT("info",
 	       ("unsuccessful connect tReturnCode %d, tNdbCon->Status() %d",
 		tReturnCode, tNdbCon->Status()));
+    releaseNdbCon(tNdbCon);
     if (theError.code == 299 || // single user mode
         theError.code == 281 )  // cluster shutdown in progress
     {
       // no need to retry with other node
       DBUG_RETURN(-1);
     }
+
+    /**
+     * If node was dead, report 0...
+     *
+     * Btw, the sendRecSignal-method should taken out and shot
+     */
+    switch(tReturnCode){
+    case -2:
+    case -3:
+      DBUG_RETURN(0);
+    }
+
     DBUG_RETURN(3);
   }//if
 }//Ndb::NDB_connect()
@@ -300,11 +316,12 @@ Ndb::waitUntilReady(int timeout)
 }
 
 /*****************************************************************************
-NdbTransaction* startTransaction();
+NdbTransaction* computeHash()
 
-Return Value:   Returns a pointer to a connection object.
-                Return NULL otherwise.
-Remark:         Start transaction. Synchronous.
+Return Value:   Returns 0 for success, NDBAPI error code otherwise
+Remark:         Computes the distribution hash value for a row with the
+                supplied distribtion key values.
+                Only relevant for natively partitioned tables.
 *****************************************************************************/ 
 int
 Ndb::computeHash(Uint32 *retval,
@@ -321,6 +338,15 @@ Ndb::computeHash(Uint32 *retval,
 
   Uint32 colcnt = impl->m_columns.size();
   Uint32 parts = impl->m_noOfDistributionKeys;
+
+  if (unlikely(impl->m_fragmentType == NdbDictionary::Object::UserDefined))
+  {
+    /* Calculating native hash on keys in user defined 
+     * partitioned table is probably part of a bug
+     */
+    goto euserdeftable;
+  }
+
   if (parts == 0)
   {
     parts = impl->m_noOfKeys;
@@ -455,6 +481,9 @@ Ndb::computeHash(Uint32 *retval,
   
   return 0;
   
+euserdeftable:
+  return 4544;
+  
 enullptr:
   return 4316;
   
@@ -550,7 +579,8 @@ Ndb::startTransaction(const NdbDictionary::Table *table,
       Uint32 hashValue;
       {
 	Uint32 buf[4];
-        Uint64 tmp[1000];
+        const Uint32 MaxKeySizeInLongWords= (NDB_MAX_KEY_SIZE + 7) / 8;
+        Uint64 tmp[ MaxKeySizeInLongWords ];
 
         if (keyLen >= sizeof(tmp))
         {
@@ -686,7 +716,7 @@ Ndb::startTransactionLocal(Uint32 aPriority, Uint32 nodeId)
   }//if
 #ifdef VM_TRACE
   if (tConnection->theListState != NdbTransaction::NotInList) {
-    printState("startTransactionLocal %x", tConnection);
+    printState("startTransactionLocal %lx", (long)tConnection);
     abort();
   }
 #endif
@@ -776,11 +806,22 @@ Ndb::closeTransaction(NdbTransaction* aConnection)
     DBUG_VOID_RETURN;
   }
   
-  if (aConnection->theReleaseOnClose == false) {
+  /**
+   * NOTE: It's ok to call getNodeSequence() here wo/ having mutex,
+   */
+  Uint32 nodeId = aConnection->getConnectedNodeId();
+  TransporterFacade* tp = theImpl->m_transporter_facade;   
+  Uint32 seq = tp->getNodeSequence(nodeId);
+  if (aConnection->theNodeSequence != seq)
+  {
+    aConnection->theReleaseOnClose = true;
+  }
+  
+  if (aConnection->theReleaseOnClose == false) 
+  {
     /**
      * Put it back in idle list for that node
      */
-    Uint32 nodeId = aConnection->getConnectedNodeId();
     aConnection->theNext = theConnectionArray[nodeId];
     theConnectionArray[nodeId] = aConnection;
     DBUG_VOID_RETURN;
@@ -1407,7 +1448,10 @@ Ndb::opTupleIdOnNdb(const NdbTableImpl* table,
   if (initAutoIncrement() == -1)
     goto error_handler;
 
-  tConnection = this->startTransaction();
+  // Start transaction with table id as hint
+  tConnection = this->startTransaction(m_sys_tab_0,
+                                       (const char *) &aTableId,
+                                       sizeof(Uint32));
   if (tConnection == NULL)
     goto error_handler;
 
@@ -1604,6 +1648,20 @@ int Ndb::setDatabaseAndSchemaName(const NdbDictionary::Table* t)
         setDatabaseName(buf);
         sprintf(buf, "%.*s", (int) (s2 - (s1 + 1)), s1 + 1);
         setDatabaseSchemaName(buf);
+#ifdef VM_TRACE
+        // verify that m_prefix looks like abc/def/
+        const char* s0 = theImpl->m_prefix.c_str();
+        const char* s1 = s0 ? strchr(s0, table_name_separator) : 0;
+        const char* s2 = s1 ? strchr(s1 + 1, table_name_separator) : 0;
+        if (!(s1 && s1 != s0 && s2 && s2 != s1 + 1 && *(s2 + 1) == 0))
+        {
+          ndbout_c("t->m_impl.m_internalName.c_str(): %s", t->m_impl.m_internalName.c_str());
+          ndbout_c("s0: %s", s0);
+          ndbout_c("s1: %s", s1);
+          ndbout_c("s2: %s", s2);
+          assert(s1 && s1 != s0 && s2 && s2 != s1 + 1 && *(s2 + 1) == 0);
+        }
+#endif
         return 0;
       }
     }
@@ -1680,7 +1738,13 @@ Ndb::internalize_table_name(const char *external_name) const
     const char* s0 = theImpl->m_prefix.c_str();
     const char* s1 = s0 ? strchr(s0, table_name_separator) : 0;
     const char* s2 = s1 ? strchr(s1 + 1, table_name_separator) : 0;
-    assert(s1 && s1 != s0 && s2 && s2 != s1 + 1 && *(s2 + 1) == 0);
+    if (!(s1 && s1 != s0 && s2 && s2 != s1 + 1 && *(s2 + 1) == 0))
+    {
+      ndbout_c("s0: %s", s0);
+      ndbout_c("s1: %s", s1);
+      ndbout_c("s2: %s", s2);
+      assert(s1 && s1 != s0 && s2 && s2 != s1 + 1 && *(s2 + 1) == 0);
+    }
 #endif
     ret.assfmt("%s%s",
                theImpl->m_prefix.c_str(),
@@ -1825,8 +1889,6 @@ int Ndb::dropEventOperation(NdbEventOperation* tOp)
   DBUG_ENTER("Ndb::dropEventOperation");
   DBUG_PRINT("info", ("name: %s", tOp->getEvent()->getTable()->getName()));
   // remove it from list
-  NdbEventOperationImpl *op=
-    NdbEventBuffer::getEventOperationImpl(tOp);
   
   theEventBuffer->dropEventOperation(tOp);
   DBUG_RETURN(0);
@@ -1893,11 +1955,9 @@ Uint64 Ndb::getLatestGCI()
 
 void Ndb::setReportThreshEventGCISlip(unsigned thresh)
 {
- if (theEventBuffer->m_free_thresh != thresh)
+ if (theEventBuffer->m_gci_slip_thresh != thresh)
  {
-   theEventBuffer->m_free_thresh= thresh;
-   theEventBuffer->m_min_free_thresh= thresh;
-   theEventBuffer->m_max_free_thresh= 100;
+   theEventBuffer->m_gci_slip_thresh= thresh;
  }
 }
 
@@ -1969,6 +2029,165 @@ Ndb::printState(const char* fmt, ...)
     theCompletedTransactionsArray[i]->printState();
   NdbMutex_Unlock(ndb_print_state_mutex);
 }
+
 #endif
 
+const char*
+Ndb::getNdbErrorDetail(const NdbError& err, char* buff, Uint32 buffLen) const
+{
+  DBUG_ENTER("Ndb::getNdbErrorDetail");
+  /* If err has non-null details member, prepare a string containing
+   * those details
+   */
+  if (!buff)
+    DBUG_RETURN(NULL);
 
+  if (err.details != NULL)
+  {
+    DBUG_PRINT("info", ("err.code is %u", err.code));
+    switch (err.code) {
+    case 893: /* Unique constraint violation */
+    {
+      /* err.details contains the violated Index's object id
+       * We'll map it to a name, then map the name to a 
+       * base table, schema and database, and put that in
+       * string form into the caller's buffer
+       */
+      UintPtr uip = (UintPtr) err.details;
+      Uint32 indexObjectId = (Uint32) (uip - (UintPtr(0)));
+      Uint32 primTableObjectId = ~ (Uint32) 0;
+      BaseString indexName;
+      char splitChars[2] = {table_name_separator, 0};
+      BaseString splitString(&splitChars[0]);
+      
+      {
+        DBUG_PRINT("info", ("Index object id is %u", indexObjectId));
+        NdbDictionary::Dictionary::List allIndices;
+        int rc = theDictionary->listObjects(allIndices, 
+                                            NdbDictionary::Object::UniqueHashIndex,
+                                            false); // FullyQualified names
+        if (rc)
+        {
+          DBUG_PRINT("info", ("listObjects call 1 failed with rc %u", rc));
+          DBUG_RETURN(NULL);
+        }
+        
+        DBUG_PRINT("info", ("Retrieved details for %u indices", allIndices.count));
+        
+        for (unsigned i = 0; i < allIndices.count; i++)
+        {
+          if (allIndices.elements[i].id == indexObjectId)
+          {
+            /* Found the index in question
+             * Expect fully qualified name to be in the form :
+             * <db>/<schema>/<primTabId>/<IndexName>
+             */
+            Vector<BaseString> idxNameComponents;
+            BaseString idxName(allIndices.elements[i].name);
+            
+            Uint32 components = idxName.split(idxNameComponents,
+                                              splitString);
+            
+            assert(components == 4);
+            
+            primTableObjectId = atoi(idxNameComponents[2].c_str());
+            indexName = idxNameComponents[3];
+            
+            DBUG_PRINT("info", ("Found index name : %s, primary table id : %u",
+                                indexName.c_str(), primTableObjectId));
+            
+            break;
+          }
+        }
+      }
+
+      if (primTableObjectId != (~(Uint32) 0))
+      {
+        NdbDictionary::Dictionary::List allTables;
+        int rc = theDictionary->listObjects(allTables,
+                                            NdbDictionary::Object::UserTable,
+                                            false); // FullyQualified names
+        
+        if (rc)
+        {
+          DBUG_PRINT("info", ("listObjects call 2 failed with rc %u", rc));
+          DBUG_RETURN(NULL);
+        }
+
+        DBUG_PRINT("info", ("Retrieved details for %u tables", allTables.count));
+
+        for (Uint32 t = 0; t < allTables.count; t++)
+        {
+          
+          if (allTables.elements[t].id == primTableObjectId)
+          {
+            /* Found table, name should be in format :
+             * <db>/<schema>/<tablename>
+             */
+            Vector<BaseString> tabNameComponents;
+            BaseString tabName(allTables.elements[t].name);
+            
+            Uint32 components = tabName.split(tabNameComponents,
+                                              splitString);
+            assert (components == 3);
+            
+            /* Now we generate a string of the format
+             * <dbname>/<schemaname>/<tabname>/<idxname>
+             * which should be usable by end users
+             */
+            BaseString result;
+            result.assfmt("%s/%s/%s/%s",
+                          tabNameComponents[0].c_str(),
+                          tabNameComponents[1].c_str(),
+                          tabNameComponents[2].c_str(),
+                          indexName.c_str());
+            
+            DBUG_PRINT("info", ("Found full index details : %s",
+                                result.c_str()));
+            
+            memcpy(buff, result.c_str(), 
+                   MIN(buffLen, 
+                       (result.length() + 1)));
+            buff[buffLen] = 0;
+
+            DBUG_RETURN(buff);
+          }
+        }
+
+        /* Primary table not found!  
+         * Strange - perhaps it's been dropped?
+         */          
+        DBUG_PRINT("info", ("Table id %u not found", primTableObjectId));
+        DBUG_RETURN(NULL);
+      }
+      else
+      {
+        /* Index not found from id - strange.
+         * Perhaps it has been dropped?
+         */
+        DBUG_PRINT("info", ("Index id %u not found", indexObjectId));
+        DBUG_RETURN(NULL);
+      }
+    }
+    default:
+    {
+      /* Unhandled details type */
+    }
+    }
+  }
+  
+  DBUG_PRINT("info", ("No details string for this error"));
+  DBUG_RETURN(NULL);
+}
+
+void
+Ndb::setCustomData(void* _customDataPtr)
+{
+  theImpl->customDataPtr = _customDataPtr;
+}
+
+void*
+Ndb::getCustomData() const
+{
+  return theImpl->customDataPtr;
+}
