@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2004, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 
 #ifndef CLUSTER_CONNECTION_IMPL_HPP
@@ -44,6 +46,7 @@ class Ndb_cluster_connection_impl : public Ndb_cluster_connection
 
   void init_get_next_node(Ndb_cluster_connection_node_iter &iter);
   Uint32 get_next_node(Ndb_cluster_connection_node_iter &iter);
+  Uint32 get_next_alive_node(Ndb_cluster_connection_node_iter &iter);
 
   inline unsigned get_connect_count() const;
 public:
@@ -90,6 +93,9 @@ private:
   Ndb* m_first_ndb_object;
   void link_ndb_object(Ndb*);
   void unlink_ndb_object(Ndb*);
+
+  BaseString m_latest_error_msg;
+  unsigned m_latest_error;
 };
 
 #endif
