@@ -1,4 +1,5 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003, 2010, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +12,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 
 /* ***************************************************
@@ -69,6 +71,7 @@ Arguments:
 extern "C" { static void* flexBenchThread(void*); }
 static int readArguments(int argc, const char** argv);
 static int createTables(Ndb*);
+static int dropTables(Ndb*);
 static void sleepBeforeStartingTest(int seconds);
 static void input_error();
 
@@ -364,7 +367,7 @@ NDB_COMMAND(flexBench, "flexBench", "flexBench", "flexbench", 65535)
      ****************************************************************/
     resetThreads(pThreadsData);
     
-    for (i = 0; i < tNoOfThreads; i++){  
+    for (i = 0; i < (int)tNoOfThreads; i++){  
       pThreadsData[i].threadNo = i;
       pThreadsData[i].threadLife = NdbThread_Create(flexBenchThread,
                                                     (void**)&pThreadsData[i],
@@ -540,7 +543,7 @@ NDB_COMMAND(flexBench, "flexBench", "flexBench", "flexbench", 65535)
     waitForThreads(pThreadsData);
 
     void * tmp;
-    for(i = 0; i<tNoOfThreads; i++){
+    for(i = 0; i<(int)tNoOfThreads; i++){
       NdbThread_WaitFor(pThreadsData[i].threadLife, &tmp);
       NdbThread_Destroy(&pThreadsData[i].threadLife);
     }
@@ -549,11 +552,13 @@ NDB_COMMAND(flexBench, "flexBench", "flexBench", "flexbench", 65535)
   if (useLongKeys == true) {
     // Only free these areas if they have been allocated
     // Otherwise cores will happen
-    for (i = 0; i < tNoOfLongPK; i++)
+    for (i = 0; i < (int)tNoOfLongPK; i++)
       free(longKeyAttrName[i]);
     free(longKeyAttrName);
   } // if
 
+  dropTables(pNdb);
+
   delete [] pThreadsData;
   delete pNdb;
   theErrorData.printErrorCounters();
@@ -588,10 +593,10 @@ static void* flexBenchThread(void* pArg)
   unsigned int      threadNo, threadBase;
   Ndb*              pNdb = NULL ;
   NdbConnection     *pTrans = NULL ;
-  NdbOperation**    pOps = NULL ;
+  const NdbOperation**    
+                    pOps = NULL ;
   StartType         tType ;
   StartType         tSaveType ;
-  NdbRecAttr*       tTmp = NULL ;
   int*              attrValue = NULL ;
   int*              attrRefValue = NULL ;
   int               check = 0 ;
@@ -620,7 +625,7 @@ static void* flexBenchThread(void* pArg)
 
   attrValue = (int*)malloc(nReadBuffSize) ;
   attrRefValue = (int*)malloc(nRefBuffSize) ;
-  pOps = (NdbOperation**)malloc(tNoOfTables*sizeof(NdbOperation*)) ;
+  pOps = (const NdbOperation**)malloc(tNoOfTables*sizeof(NdbOperation*)) ;
   pNdb = new Ndb(g_cluster_connection, "TEST_DB" );
   pRec= (NdbRecord **)calloc(tNoOfTables*3, sizeof(*pRec));
   pAttrSet= (unsigned char **)calloc(tNoOfTables, sizeof(*pAttrSet));
@@ -644,7 +649,7 @@ static void* flexBenchThread(void* pArg)
   
   /* Set up NdbRecord's for the tables. */
   dict= pNdb->getDictionary();
-  for (int tab= 0; tab<tNoOfTables; tab++)
+  for (int tab= 0; tab<(int)tNoOfTables; tab++)
   {
     const NdbDictionary::Table *table= dict->getTable(tableName[tab]);
     int numPKs= (useLongKeys ? tNoOfLongPK : 1);
@@ -770,7 +775,7 @@ static void* flexBenchThread(void* pArg)
     // Calculate offset value before going into the next loop
     nRefOpOffset = tAttributeSize*tNoOfAttributes*(ops-1) ; 
     for(Uint32 a = 0 ; a < tNoOfAttributes ; a++)
-      for(Uint32 b= 0; b<tAttributeSize; b++)
+      for(Uint32 b= 0; b<(Uint32)tAttributeSize; b++)
         attrRefValue[nRefOpOffset + tAttributeSize*a + b] = 
           (int)(threadBase + ops + a) ;
   }
@@ -858,7 +863,6 @@ static void* flexBenchThread(void* pArg)
         const NdbRecord *pk_record= pRec[countTables];
         const NdbRecord *attr_record= pRec[countTables+tNoOfTables];
         const NdbRecord *all_record= pRec[countTables+2*tNoOfTables];
-        const char *tabName= tableName[countTables];
 
 	switch (tType) {
 	case stInsert:          // Insert case
@@ -887,7 +891,8 @@ static void* flexBenchThread(void* pArg)
                                                    attr_record, pRowAttr);
 	  break;
 	case stDelete:          // Delete Case
-	  pOps[countTables]= pTrans->deleteTuple(pk_record, pRowPK);
+	  pOps[countTables]= pTrans->deleteTuple(pk_record, pRowPK,
+                                                 attr_record);
 	  break;
 	case stVerify:
 	  pOps[countTables]= pTrans->readTuple(pk_record, pRowPK,
@@ -1195,13 +1200,13 @@ static void sleepBeforeStartingTest(int seconds){
 static int
 createTables(Ndb* pMyNdb){
   int i;
-  for (i = 0; i < tNoOfAttributes; i++){
+  for (i = 0; i < (int)tNoOfAttributes; i++){
     BaseString::snprintf(attrName[i], MAXSTRLEN, "COL%d", i);
   }
 
   // Note! Uses only uppercase letters in table name's
   // so that we can look at the tables with SQL
-  for (i = 0; i < tNoOfTables; i++){
+  for (i = 0; i < (int)tNoOfTables; i++){
     if (theStdTableNameFlag == 0){
       BaseString::snprintf(tableName[i], MAXSTRLEN, "TAB%d_%d", i, 
 	       (int)(NdbTick_CurrentMillisecond() / 1000));
@@ -1210,7 +1215,7 @@ createTables(Ndb* pMyNdb){
     }
   }
   
-  for(i = 0; i < tNoOfTables; i++){
+  for(i = 0; i < (int)tNoOfTables; i++){
     ndbout << "Creating " << tableName[i] << "... ";
     
     NdbDictionary::Table tmpTable(tableName[i]);
@@ -1251,6 +1256,21 @@ createTables(Ndb* pMyNdb){
   return 0;
 }
 
+static int
+dropTables(Ndb* pMyNdb){
+  unsigned int i;
+
+  // Note! Uses only uppercase letters in table name's
+  // so that we can look at the tables with SQL
+  for(i = 0; i < tNoOfTables; i++){
+    ndbout << "Dropping " << tableName[i] << "... ";
+    pMyNdb->getDictionary()->dropTable(tableName[i]);
+    ndbout << "done" << endl;
+  }
+  
+  return 0;
+}
+
       
 static void input_error(){
   ndbout << endl << "Invalid argument!" << endl;
