@@ -1,4 +1,6 @@
-/* Copyright (C) 2003 MySQL AB
+/*
+   Copyright (c) 2003-2008 MySQL AB, 2008, 2009 Sun Microsystems, Inc.
+   Use is subject to license terms.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -11,7 +13,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 #include "HugoTransactions.hpp"
 #include <NDBT_Stats.hpp>
@@ -51,8 +54,8 @@ HugoTransactions::scanReadRecords(Ndb* pNdb,
   while (true){
 
     if (retryAttempt >= m_retryMax){
-      g_err << "ERROR: has retried this operation " << retryAttempt 
-	    << " times, failing!" << endl;
+      g_err << __LINE__ << " ERROR: has retried this operation " 
+            << retryAttempt << " times, failing!" << endl;
       return NDBT_FAILED;
     }
 
@@ -83,13 +86,6 @@ HugoTransactions::scanReadRecords(Ndb* pNdb,
       return NDBT_FAILED;
     }
     
-    check = pOp->interpret_exit_ok();
-    if( check == -1 ) {
-      ERR(pTrans->getNdbError());
-      closeTransaction(pNdb);
-      return NDBT_FAILED;
-    }
-  
     for(a = 0; a<tab.getNoOfColumns(); a++){
       if((row.attributeStore(a) = 
 	  pOp->getValue(tab.getColumn(a)->getName())) == 0) {
@@ -149,7 +145,7 @@ HugoTransactions::scanReadRecords(Ndb* pNdb,
     }
     if (eof == -1) {
       const NdbError err = pTrans->getNdbError();
-      
+
       if (err.status == NdbError::TemporaryError){
 	ERR_INFO(err);
 	closeTransaction(pNdb);
@@ -161,6 +157,18 @@ HugoTransactions::scanReadRecords(Ndb* pNdb,
 	  // Too many active scans, no limit on number of retry attempts
 	  break;
 	default:
+          if (err.classification == NdbError::TimeoutExpired)
+          {
+            if (retryAttempt >= (m_retryMax / 10) && 
+                (parallelism == 0 || parallelism > 1))
+            {
+              /**
+               * decrease parallelism
+               */
+              parallelism = 1;
+              ndbout_c("decrease parallelism");
+            }
+          }
 	  retryAttempt++;
 	}
 	continue;
@@ -202,8 +210,10 @@ HugoTransactions::scanReadRecords(Ndb* pNdb,
   while (true){
 
     if (retryAttempt >= m_retryMax){
-      g_err << "ERROR: has retried this operation " << retryAttempt 
-	    << " times, failing!" << endl;
+      g_err << __LINE__ << " ERROR: has retried this operation " 
+            << retryAttempt  << " times, failing!" << endl;
+      g_err << "lm: " << Uint32(lm) << " flags: H'" << hex << scan_flags
+            << endl;
       return NDBT_FAILED;
     }
 
@@ -234,13 +244,6 @@ HugoTransactions::scanReadRecords(Ndb* pNdb,
       return NDBT_FAILED;
     }
     
-    check = pOp->interpret_exit_ok();
-    if( check == -1 ) {
-      ERR(pTrans->getNdbError());
-      closeTransaction(pNdb);
-      return NDBT_FAILED;
-    }
-  
     for(a = 0; a<tab.getNoOfColumns(); a++){
       if((row.attributeStore(a) = 
 	  pOp->getValue(tab.getColumn(a)->getName())) == 0) {
@@ -312,6 +315,31 @@ HugoTransactions::scanReadRecords(Ndb* pNdb,
 	  // Too many active scans, no limit on number of retry attempts
 	  break;
 	default:
+          if (err.classification == NdbError::TimeoutExpired)
+          {
+            if (retryAttempt >= (m_retryMax / 10) && 
+                (parallelism == 0 || parallelism > 1))
+            {
+              /**
+               * decrease parallelism
+               */
+              parallelism = 1;
+              ndbout_c("decrease parallelism");
+            }
+            else if (retryAttempt >= (m_retryMax / 5) &&
+                     (lm != NdbOperation::LM_CommittedRead))
+            {
+              lm = NdbOperation::LM_CommittedRead;
+              ndbout_c("switch to LM_CommittedRead");
+            }
+            else if (retryAttempt >= (m_retryMax / 4) &&
+                     (pIdx != 0))
+            {
+              pIdx = 0;
+              scan_flags |= NdbScanOperation::SF_TupScan;
+              ndbout_c("switch to table-scan (SF_TupScan) form index-scan");
+            }
+          }
 	  retryAttempt++;
 	}
 	continue;
@@ -757,8 +785,8 @@ HugoTransactions::fillTableStartFrom(Ndb* pNdb,
     
     // Execute the transaction and insert the record
     check = pTrans->execute(Commit, CommitAsMuchAsPossible); 
-    if(check == -1 ) {
-      const NdbError err = pTrans->getNdbError();
+    const NdbError err = pTrans->getNdbError();
+    if(check == -1 || err.code != 0) {
       closeTransaction(pNdb);
       
       switch(err.status){
@@ -826,7 +854,8 @@ int
 HugoTransactions::pkReadRecords(Ndb* pNdb, 
 				int records,
 				int batch,
-				NdbOperation::LockMode lm){
+				NdbOperation::LockMode lm,
+                                int _rand){
   int                  reads = 0;
   int                  r = 0;
   int                  retryAttempt = 0;
@@ -871,14 +900,41 @@ HugoTransactions::pkReadRecords(Ndb* pNdb,
     if (timer_active)
       NdbTick_getMicroTimer(&timer_start);
 
-    if(pkReadRecord(pNdb, r, batch, lm) != NDBT_OK)
+    NdbOperation::LockMode lmused;
+    if (_rand == 0)
     {
-      ERR(pTrans->getNdbError());
-      closeTransaction(pNdb);
-      return NDBT_FAILED;
+      if(pkReadRecord(pNdb, r, batch, lm, &lmused) != NDBT_OK)
+      {
+        ERR(pTrans->getNdbError());
+        closeTransaction(pNdb);
+        return NDBT_FAILED;
+      }
+    }
+    else
+    {
+      if(pkReadRandRecord(pNdb, records, batch, lm, &lmused) != NDBT_OK)
+      {
+        ERR(pTrans->getNdbError());
+        closeTransaction(pNdb);
+        return NDBT_FAILED;
+      }
     }
     
-    check = pTrans->execute(Commit, AbortOnError);   
+    check = pTrans->execute(Commit, AbortOnError);
+
+    if (check != -1 && lmused == NdbOperation::LM_CommittedRead)
+    {
+      /**
+       * LM_CommittedRead will not abort transaction
+       *   even if doing execute(AbortOnError);
+       *   so also check pTrans->getNdbError() in this case
+       */
+      if (pTrans->getNdbError().status != NdbError::Success)
+      {
+        check = -1;
+      }
+    }      
+
     if( check == -1 ) {
       const NdbError err = pTrans->getNdbError();
       
@@ -902,17 +958,21 @@ HugoTransactions::pkReadRecords(Ndb* pNdb,
       }
     } else {
 
-      if(pIndexScanOp)
+      if(indexScans.size() > 0)
       {
+        /* Index scan used to read records....*/
 	int rows_found = 0;
-	while((check = pIndexScanOp->nextResult()) == 0)
-	{
-	  rows_found++;
-	  if (calc.verifyRowValues(rows[0]) != 0){
-	    closeTransaction(pNdb);
-	    return NDBT_FAILED;
-	  }
-	}
+        for (Uint32 scanOp=0; scanOp < indexScans.size(); scanOp++)
+        {
+          while((check = indexScans[scanOp]->nextResult()) == 0)
+          {
+            rows_found++;
+            if (calc.verifyRowValues(rows[0]) != 0){
+              closeTransaction(pNdb);
+              return NDBT_FAILED;
+            }
+          }
+        }
 	if(check != 1 || rows_found > batch)
 	{
 	  closeTransaction(pNdb);
@@ -951,6 +1011,7 @@ HugoTransactions::pkReadRecords(Ndb* pNdb,
     }
   }
   deallocRows();
+  indexScans.clear();
   g_info << reads << " records read" << endl;
   return NDBT_OK;
 }
@@ -1038,38 +1099,59 @@ HugoTransactions::pkUpdateRecords(Ndb* pNdb,
     if (timer_active)
       NdbTick_getMicroTimer(&timer_start);
 
-    if(pIndexScanOp)
+    int rows_found = 0;
+
+    if(indexScans.size() > 0)
     {
-      int rows_found = 0;
-      while((check = pIndexScanOp->nextResult(true)) == 0)
+      /* Index scans used to read records */
+      for (Uint32 scanOp=0; scanOp < indexScans.size(); scanOp++)
       {
-	do {
-	  
-	  if (calc.verifyRowValues(rows[0]) != 0){
-	    closeTransaction(pNdb);
-	    return NDBT_FAILED;
-	  }
-	  
-	  int updates = calc.getUpdatesValue(rows[0]) + 1;
-	  
-	  if(pkUpdateRecord(pNdb, r+rows_found, 1, updates) != NDBT_OK)
-	  {
-	    ERR(pTrans->getNdbError());
-	    closeTransaction(pNdb);
-	    return NDBT_FAILED;
-	  }
-	  rows_found++;
-	} while((check = pIndexScanOp->nextResult(false)) == 0);
-	
-	if(check != 2)
-	  break;
-	if((check = pTrans->execute(NoCommit, AbortOnError)) != 0)
-	  break;
-      }
-      if(check != 1 || rows_found != batch)
+        while((check = indexScans[scanOp]->nextResult(true)) == 0)
+        {
+          do {
+            
+            if (calc.verifyRowValues(rows[0]) != 0){
+              g_info << "Row validation failure" << endl;
+              closeTransaction(pNdb);
+              return NDBT_FAILED;
+            }
+            
+            int updates = calc.getUpdatesValue(rows[0]) + 1;
+            
+            /* Rows may not arrive in the order they were requested
+             * (When multiple partitions scanned without ORDERBY)
+             * therefore we use the id from the row to update it
+             */
+            const Uint32 rowId= calc.getIdValue(rows[0]);
+            if(pkUpdateRecord(pNdb, rowId, 1, updates) != NDBT_OK)
+            {
+              ERR(pTrans->getNdbError());
+              closeTransaction(pNdb);
+              return NDBT_FAILED;
+            }
+            rows_found++;
+          } while((check = indexScans[scanOp]->nextResult(false)) == 0);
+          
+          if(check != 2)
+            break;
+          if((check = pTrans->execute(NoCommit, AbortOnError)) != 0)
+            break;
+        } // Next fetch on this scan op...
+
+        if(check != 1)
+        {
+          g_info << "Check failed" << endl;
+          closeTransaction(pNdb);
+          return NDBT_FAILED;
+        } 
+      } // Next scan op...
+
+      if (rows_found != batch)
       {
-	closeTransaction(pNdb);
-	return NDBT_FAILED;
+        g_info << "Incorrect num of rows found.  Expected "
+               << batch << ". Found " << rows_found << endl;
+        closeTransaction(pNdb);
+        return NDBT_FAILED;
       }
     }
     else
@@ -1126,6 +1208,7 @@ HugoTransactions::pkUpdateRecords(Ndb* pNdb,
   }
   
   deallocRows();
+  indexScans.clear();
   g_info << "|- " << updated << " records updated" << endl;
   return NDBT_OK;
 }
@@ -1866,3 +1949,4 @@ HugoTransactions::indexUpdateRecords(Ndb* pNdb,
 }
 
 template class Vector<NDBT_ResultRow*>;
+template class Vector<NdbIndexScanOperation*>;
